# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper for B |eacute| zier Curves.

.. |eacute| unicode:: U+000E9 .. LATIN SMALL LETTER E WITH ACUTE
   :trim:

See :doc:`../curve-curve-intersection` for examples using the
:class:`Curve` class to find intersections.

.. testsetup:: *

   import numpy as np
   import bezier
"""


import matplotlib.pyplot as plt
import numpy as np

from bezier import _base
from bezier import _curve_helpers
from bezier import _intersection_helpers


_REPR_TEMPLATE = (
    '<{} (degree={:d}, dimension={:d}, start={:g}, end={:g})>')
_LINEAR_SUBDIVIDE = np.array([
    [1.0, 0.0],
    [0.5, 0.5],
    [0.0, 1.0],
])
_QUADRATIC_SUBDIVIDE = np.array([
    [1.0, 0.0, 0.0],
    [0.5, 0.5, 0.0],
    [0.25, 0.5, 0.25],
    [0.0, 0.5, 0.5],
    [0.0, 0.0, 1.0],
])
_CUBIC_SUBDIVIDE = np.array([
    [1.0, 0.0, 0.0, 0.0],
    [0.5, 0.5, 0.0, 0.0],
    [0.25, 0.5, 0.25, 0.0],
    [0.125, 0.375, 0.375, 0.125],
    [0.0, 0.25, 0.5, 0.25],
    [0.0, 0.0, 0.5, 0.5],
    [0.0, 0.0, 0.0, 1.0],
])


class Curve(_base.Base):
    r"""Represents a B |eacute| zier `curve`_.

    .. _curve: https://en.wikipedia.org/wiki/B%C3%A9zier_curve

    We take the traditional definition: a B |eacute| zier curve is a mapping
    from :math:`s \in \left[0, 1\right]` to convex combinations
    of points :math:`v_0, v_1, \ldots, v_n` in some vector space:

    .. math::

       B(s) = \sum_{j = 0}^n \binom{n}{j} s^j (1 - s)^{n - j} \cdot v_j

    .. image:: ../images/curve_constructor.png
       :align: center

    .. doctest:: curve-constructor

       >>> import bezier
       >>> nodes = np.array([
       ...     [0.0  , 0.0],
       ...     [0.625, 0.5],
       ...     [1.0  , 0.5],
       ... ])
       >>> curve = bezier.Curve(nodes)
       >>> curve
       <Curve (degree=2, dimension=2)>

    .. testcleanup:: curve-constructor

       import make_images
       make_images.curve_constructor(curve)

    Args:
        nodes (numpy.ndarray): The nodes in the curve. The rows
            represent each node while the columns are the dimension
            of the ambient space.
        start (Optional[float]): The beginning of the sub-interval
            that this curve represents.
        end (Optional[float]): The end of the sub-interval
            that this curve represents.
        root (Optional[Curve]): The root curve that contains this
            current curve.
        _copy (bool): Flag indicating if the nodes should be copied before
            being stored. Defaults to :data:`True` since callers may
            freely mutate ``nodes`` after passing in.
    """

    _length = None
    _edge_index = None
    _next_edge = None
    _previous_edge = None

    def __init__(self, nodes, start=0.0, end=1.0, root=None, _copy=True):
        super(Curve, self).__init__(nodes, _copy=_copy)
        self._start = start
        self._end = end
        if root is None:
            root = self
        self._root = root

    def __repr__(self):
        """Representation of current object.

        Returns:
            str: Object representation.
        """
        if self.start == 0.0 and self.end == 1.0:
            return super(Curve, self).__repr__()
        else:
            return _REPR_TEMPLATE.format(
                self.__class__.__name__, self.degree,
                self.dimension, self.start, self.end)

    @staticmethod
    def _get_degree(num_nodes):
        """Get the degree of the current curve.

        Args:
            num_nodes (int): The number of nodes provided.

        Returns:
            int: The degree of the current curve.
        """
        return num_nodes - 1

    @property
    def length(self):
        """float: The length of the current curve."""
        if self._length is None:
            self._length = _curve_helpers.compute_length(
                self._nodes, self.degree)
        return self._length

    @property
    def start(self):
        """float: Start of sub-interval this curve represents.

        This value is used to track the current curve in the
        re-parameterization / subdivision process. The curve is still
        defined on the unit interval, but this value illustrates
        how this curve relates to a "parent" curve. For example:

        .. doctest:: curve-start
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.array([
           ...     [0.0, 0.0],
           ...     [1.0, 2.0],
           ... ])
           >>> curve = bezier.Curve(nodes)
           >>> curve
           <Curve (degree=1, dimension=2)>
           >>> left, right = curve.subdivide()
           >>> left
           <Curve (degree=1, dimension=2, start=0, end=0.5)>
           >>> right
           <Curve (degree=1, dimension=2, start=0.5, end=1)>
           >>> _, mid_right = left.subdivide()
           >>> mid_right
           <Curve (degree=1, dimension=2, start=0.25, end=0.5)>
           >>> mid_right.nodes
           array([[ 0.25, 0.5 ],
                  [ 0.5 , 1.  ]])
        """
        return self._start

    @property
    def end(self):
        """float: End of sub-interval this curve represents.

        See :attr:`~Curve.start` for more information.
        """
        return self._end

    @property
    def root(self):
        """Curve: The "root" curve that contains the current curve.

        This indicates that the current curve is a section of the
        "root" curve. For example:

        .. testsetup:: curve-root

           import numpy as np
           import bezier

           nodes = np.array([
               [0.0, 0.0],
               [0.75, 0.0],
               [1.0, 1.0],
           ])
           curve = bezier.Curve(nodes)

        .. doctest:: curve-root
           :options: +NORMALIZE_WHITESPACE

           >>> _, right = curve.subdivide()
           >>> right
           <Curve (degree=2, dimension=2, start=0.5, end=1)>
           >>> right.root is curve
           True
           >>> right.evaluate(0.0) == curve.evaluate(0.5)
           array([ True, True], dtype=bool)
           >>>
           >>> mid_left, _ = right.subdivide()
           >>> mid_left
           <Curve (degree=2, dimension=2, start=0.5, end=0.75)>
           >>> mid_left.root is curve
           True
           >>> mid_left.evaluate(1.0) == curve.evaluate(0.75)
           array([ True, True], dtype=bool)
        """
        return self._root

    @property
    def edge_index(self):
        """Optional[int]: The index of the edge among a group of edges.

        .. testsetup:: edge-index

           import numpy as np
           import bezier

           surface = bezier.Surface(np.array([
               [0.0, 0.0],
               [1.0, 0.0],
               [0.0, 1.0],
           ]))
           _, curve, _ = surface.edges

        .. doctest:: edge-index

           >>> curve.edge_index
           1
           >>> curve.previous_edge
           <Curve (degree=1, dimension=2)>
           >>> curve.previous_edge.edge_index
           0
           >>> curve.next_edge
           <Curve (degree=1, dimension=2)>
           >>> curve.next_edge.edge_index
           2

        This is intended to be used when a :class:`Curve` is created as part
        of a larger structure like a :class:`.Surface` or
        :class:`.CurvedPolygon`.
        """
        return self._edge_index

    @property
    def next_edge(self):
        """Optional[Curve]: An edge that comes after the current one.

        This is intended to be used when a :class:`Curve` is created as part
        of a larger structure like a :class:`.Surface` or
        :class:`.CurvedPolygon`.
        """
        return self._next_edge

    @property
    def previous_edge(self):
        """Optional[Curve]: An edge that comes before the current one.

        This is intended to be used when a :class:`Curve` is created as part
        of a larger structure like a :class:`.Surface` or
        :class:`.CurvedPolygon`.
        """
        return self._previous_edge

    def evaluate(self, s):
        r"""Evaluate :math:`B(s)` along the curve.

        See :meth:`evaluate_multi` for more details.

        .. image:: ../images/curve_evaluate.png
           :align: center

        .. doctest:: curve-eval
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.array([
           ...     [0.0  , 0.0],
           ...     [0.625, 0.5],
           ...     [1.0  , 0.5],
           ... ])
           >>> curve = bezier.Curve(nodes)
           >>> curve.evaluate(0.75)
           array([ 0.796875, 0.46875 ])

        .. testcleanup:: curve-eval

           import make_images
           make_images.curve_evaluate(curve)

        Args:
            s (float): Parameter along the curve.

        Returns:
            numpy.ndarray: The point on the curve (as a one dimensional
            NumPy array).
        """
        result = self.evaluate_multi(np.array([s]))
        return result.flatten()

    def evaluate_multi(self, s_vals):
        r"""Evaluate :math:`B(s)` for multiple points along the curve.

        This is done by first evaluating each member of the
        `Bernstein basis`_ at each value in ``s_vals`` and then
        applying those to the control points for the current curve.

        This is done instead of using `de Casteljau's algorithm`_.
        Implementing de Casteljau is problematic because it requires
        a choice between one of two methods:

        * vectorize operations of the form :math:`(1 - s)v + s w`,
          which requires a copy of the curve's control points for
          each value in ``s_vals``
        * avoid vectorization and compute each point in serial

        Instead, we can use vectorized operations to build up the
        Bernstein basis values.

        .. _Bernstein basis:
            https://en.wikipedia.org/wiki/Bernstein_polynomial
        .. _de Casteljau's algorithm:
            https://en.wikipedia.org/wiki/De_Casteljau%27s_algorithm

        .. doctest:: curve-eval-multi
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.array([
           ...     [0.0, 0.0, 0.0],
           ...     [1.0, 2.0, 3.0],
           ... ])
           >>> curve = bezier.Curve(nodes)
           >>> curve
           <Curve (degree=1, dimension=3)>
           >>> s_vals = np.linspace(0.0, 1.0, 5)
           >>> curve.evaluate_multi(s_vals)
           array([[ 0.  , 0.  , 0.  ],
                  [ 0.25, 0.5 , 0.75],
                  [ 0.5 , 1.  , 1.5 ],
                  [ 0.75, 1.5 , 2.25],
                  [ 1.  , 2.  , 3.  ]])

        Args:
            s_vals (numpy.ndarray): Parameters along the curve (as a
                1D array).

        Returns:
            numpy.ndarray: The points on the curve. As a two dimensional
            NumPy array, with the rows corresponding to each ``s``
            value and the columns to the dimension.
        """
        return _curve_helpers.evaluate_multi(
            self._nodes, self.degree, s_vals)

    def plot(self, num_pts, ax=None, show=False):
        """Plot the current curve.

        Args:
            num_pts (int): Number of points to plot.
            ax (Optional[matplotlib.artist.Artist]): matplotlib axis object
                to add plot to.
            show (Optional[bool]): Flag indicating if the plot should be
                shown.

        Returns:
            matplotlib.artist.Artist: The axis containing the plot. This
            may be a newly created axis.

        Raises:
            NotImplementedError: If the curve's dimension is not ``2``.
        """
        if self.dimension != 2:
            raise NotImplementedError('2D is the only supported dimension',
                                      'Current dimension', self.dimension)

        s_vals = np.linspace(0.0, 1.0, num_pts)
        points = self.evaluate_multi(s_vals)

        if ax is None:
            fig = plt.figure()
            ax = fig.gca()

        ax.plot(points[:, 0], points[:, 1])

        if show:
            plt.show()

        return ax

    def subdivide(self):
        r"""Split the curve :math:`B(s)` into a left and right half.

        Takes the interval :math:`\left[0, 1\right]` and splits the curve into
        :math:`B_1 = B\left(\left[0, \frac{1}{2}\right]\right)` and
        :math:`B_2 = B\left(\left[\frac{1}{2}, 1\right]\right)`. In
        order to do this, also reparameterizes the curve, hence the resulting
        left and right halves have new nodes.

        .. image:: ../images/curve_subdivide.png
           :align: center

        .. doctest:: curve-subdivide
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.array([
           ...     [0.0 , 0.0],
           ...     [1.25, 3.0],
           ...     [2.0 , 1.0],
           ... ])
           >>> curve = bezier.Curve(nodes)
           >>> left, right = curve.subdivide()
           >>> left
           <Curve (degree=2, dimension=2, start=0, end=0.5)>
           >>> left.nodes
           array([[ 0.   , 0.   ],
                  [ 0.625, 1.5  ],
                  [ 1.125, 1.75 ]])
           >>> right
           <Curve (degree=2, dimension=2, start=0.5, end=1)>
           >>> right.nodes
           array([[ 1.125, 1.75 ],
                  [ 1.625, 2.   ],
                  [ 2.   , 1.   ]])

        .. testcleanup:: curve-subdivide

           import make_images
           make_images.curve_subdivide(curve, left, right)

        Returns:
            Tuple[Curve, Curve]: The left and right sub-curves.
        """
        if self.degree == 1:
            # pylint: disable=no-member
            new_nodes = _LINEAR_SUBDIVIDE.dot(self._nodes)
            # pylint: enable=no-member
        elif self.degree == 2:
            # pylint: disable=no-member
            new_nodes = _QUADRATIC_SUBDIVIDE.dot(self._nodes)
            # pylint: enable=no-member
        elif self.degree == 3:
            # pylint: disable=no-member
            new_nodes = _CUBIC_SUBDIVIDE.dot(self._nodes)
            # pylint: enable=no-member
        else:
            subdivide_mat = _curve_helpers.make_subdivision_matrix(
                self.degree)
            new_nodes = subdivide_mat.dot(self._nodes)

        left_nodes = new_nodes[:self.degree + 1, :]
        right_nodes = new_nodes[self.degree:, :]

        midpoint = 0.5 * (self.start + self.end)
        left = Curve(left_nodes, start=self.start,
                     end=midpoint, root=self.root, _copy=False)
        right = Curve(right_nodes, start=midpoint,
                      end=self.end, root=self.root, _copy=False)
        return left, right

    def intersect(self, other):
        """Find the points of intersection with another curve.

        See :doc:`../curve-curve-intersection` for more details.

        .. image:: ../images/curve_intersect.png
           :align: center

        .. doctest:: curve-intersect
           :options: +NORMALIZE_WHITESPACE

           >>> nodes1 = np.array([
           ...     [0.0  , 0.0  ],
           ...     [0.375, 0.75 ],
           ...     [0.75 , 0.375],
           ... ])
           >>> curve1 = bezier.Curve(nodes1)
           >>> nodes2 = np.array([
           ...     [0.5, 0.0 ],
           ...     [0.5, 0.75],
           ... ])
           >>> curve2 = bezier.Curve(nodes2)
           >>> intersections = curve1.intersect(curve2)
           >>> intersections
           array([[ 0.5, 0.5]])

        .. testcleanup:: curve-intersect

           import make_images
           make_images.curve_intersect(curve1, curve2, intersections)

        Args:
            other (Curve): Other curve to intersect with.

        Returns:
            numpy.ndarray: Array of intersection points (possibly empty).

        Raises:
            TypeError: If ``other`` is not a curve.
            NotImplementedError: If both curves aren't two-dimensional.
        """
        if not isinstance(other, Curve):
            raise TypeError('Can only intersect with another curve',
                            'Received', other)
        if self.dimension != 2 or other.dimension != 2:
            raise NotImplementedError(
                'Intersection only implemented in 2D')

        candidates = [(self, other)]
        intersections = _intersection_helpers.all_intersections(candidates)
        if intersections:
            return np.vstack([intersection.point
                              for intersection in intersections])
        else:
            return np.zeros((0, 2))

    def elevate(self):
        r"""Return a degree-elevated version of the current curve.

        Does this by converting the current nodes :math:`v_0, \ldots, v_n`
        to new nodes :math:`w_0, \ldots, w_{n + 1}` where

        .. math::

           \begin{align*}
           w_0 &= v_0 \\
           w_j &= \frac{j}{n + 1} v_{j - 1} + \frac{n + 1 - j}{n + 1} v_j \\
           w_{n + 1} &= v_n
           \end{align*}

        Returns:
            Curve: The degree-elevated curve.
        """
        new_nodes = _curve_helpers.elevate_nodes(
            self._nodes, self.degree, self.dimension)
        return Curve(new_nodes, _copy=False)

    def specialize(self, start, end):
        """Specialize the curve to a given sub-interval.

        .. image:: ../images/curve_specialize.png
           :align: center

        .. doctest:: curve-specialize

           >>> curve = bezier.Curve(np.array([
           ...     [0.0, 0.0],
           ...     [0.5, 1.0],
           ...     [1.0, 0.0],
           ... ]))
           >>> new_curve = curve.specialize(-0.25, 0.75)
           >>> new_curve
           <Curve (degree=2, dimension=2, start=-0.25, end=0.75)>
           >>> new_curve.nodes
           array([[-0.25 , -0.625],
                  [ 0.25 ,  0.875],
                  [ 0.75 ,  0.375]])

        .. testcleanup:: curve-specialize

           import make_images
           make_images.curve_specialize(curve, new_curve)

        This is generalized version of :meth:`subdivide`, and can even
        match the output of that method:

        .. testsetup:: curve-specialize2

           import numpy as np
           import bezier

           curve = bezier.Curve(np.array([
               [0.0, 0.0],
               [0.5, 1.0],
               [1.0, 0.0],
           ]))

        .. doctest:: curve-specialize2

            >>> left, right = curve.subdivide()
            >>> left == curve.specialize(0.0, 0.5)
            True
            >>> right == curve.specialize(0.5, 1.0)
            True

        Args:
            start (float): The start point of the interval we
                are specializing to.
            end (float): The end point of the interval we
                are specializing to.

        Returns:
            Curve: The newly-specialized curve.
        """
        new_nodes = _curve_helpers.specialize_curve(
            self._nodes, self.degree, start, end)

        interval_delta = self.end - self.start
        true_start = self.start + start * interval_delta
        true_end = self.start + end * interval_delta
        return Curve(new_nodes, start=true_start,
                     end=true_end, root=self.root,
                     _copy=False)

    def locate(self, point):
        r"""Find a point on the current curve.

        Solves for :math:`s` in :math:`B(s) = p`.

        .. note::

           A unique solution is only guaranteed if the current curve has no
           self-intersections. This code assumes, but doesn't check, that
           this is true.

        .. image:: ../images/curve_locate.png
           :align: center

        .. doctest:: curve-locate

           >>> curve = bezier.Curve(np.array([
           ...     [0.0, 0.0],
           ...     [1.0, 2.0],
           ...     [3.0, 1.0],
           ...     [4.0, 0.0],
           ... ]))
           >>> point1 = np.array([[3.09375, 0.703125]])
           >>> s = curve.locate(point1)
           >>> s
           0.75
           >>> point2 = np.array([[2.0, 0.5]])
           >>> curve.locate(point2) is None
           True

        .. testcleanup:: curve-locate

           import make_images
           make_images.curve_locate(curve, point1, point2)

        Args:
            point (numpy.ndarray): A (``1xD``) point on the curve,
                where :math:`D` is the dimension of the curve.

        Returns:
            Optional[float]: The parameter value (:math:`s`) corresponding
            to ``point`` or :data:`None` if the point is not on
            the ``curve``.

        Raises:
            ValueError: If the dimension of the ``point`` doesn't match the
                dimension of the current curve.
        """
        if point.shape != (1, self.dimension):
            raise ValueError('Point is not in same dimension as curve',
                             point, 'Shape expected:', (1, self.dimension))

        return _curve_helpers.locate_point(self, point)



# See the License for the specific language governing permissions and
# limitations under the License.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest


WRONG_SHAPE_TEMPLATE = """\
Arrays have different shapes
array1{} =
{!r}
array2{} =
{!r}
"""
NOT_EQ_TEMPLATE = """\
Arrays not equal
array1 =
{!r}
array2 =
{!r}
"""


def get_random(seed):
    import numpy as np

    # pylint: disable=no-member
    return np.random.RandomState(seed=seed)
    # pylint: enable=no-member


def binary_round(value, num_bits):
    # NOTE: This assumes ``value`` is not Inf/-Inf/NaN or
    #       a subnormal number.
    hex_val = value.hex()
    pre, hex_digits = hex_val.split('0x1.')
    hex_digits, post = hex_digits.split('p')
    assert len(hex_digits) == 13

    # NOTE: Python produces 0b100101... instead of 100101...
    all_bits = bin(int(hex_digits, 16))[2:]
    all_bits = all_bits.zfill(52)
    assert len(all_bits) == 52

    truncated_bits = all_bits[:num_bits] + '0' * (52 - num_bits)
    truncated_hex = '{:013x}'.format(int(truncated_bits, 2))

    python_hex = pre + '0x1.' + truncated_hex + 'p' + post
    return float.fromhex(python_hex)


def get_random_nodes(shape, seed, num_bits):
    import functools
    import numpy as np

    random_state = get_random(seed)
    nodes = random_state.random_sample(shape)
    # Round the nodes to ``num_bits`` bits to avoid round-off.
    to_vectorize = functools.partial(binary_round, num_bits=num_bits)
    return np.vectorize(to_vectorize)(nodes)


def ref_triangle_uniform_nodes(pts_exponent):
    import numpy as np
    import six

    # Using the exponent means that we will divide by
    # 2**exp, which can be done without roundoff (for small
    # enough exponents).
    pts_per_side = 2**pts_exponent + 1
    total = ((pts_per_side + 1) * pts_per_side) // 2
    result = np.zeros((total, 2))

    index = 0
    for y_val in six.moves.xrange(pts_per_side):
        remaining = pts_per_side - y_val
        for x_val in six.moves.xrange(remaining):
            result[index, :] = x_val, y_val
            index += 1

    result /= (pts_per_side - 1.0)
    return result


def check_plot_call(test_case, call, expected, **kwargs):
    # Unpack the call as name, positional args, keyword args
    _, positional, keyword = call
    test_case.assertEqual(keyword, kwargs)
    test_case.assertEqual(len(positional), 2)
    test_case.assertEqual(positional[0], expected[:, 0])
    test_case.assertEqual(positional[1], expected[:, 1])


class NumPyTestCase(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        import numpy as np

        super(NumPyTestCase, self).__init__(*args, **kwargs)
        self.addTypeEqualityFunc(np.ndarray, self.assertArrayEqual)

    def assertArrayEqual(self, arr1, arr2, msg=None):
        import numpy as np

        if not arr1.shape == arr2.shape:  # pragma: NO COVER
            standard_msg = WRONG_SHAPE_TEMPLATE.format(
                arr1.shape, arr1, arr2.shape, arr2)
            self.fail(self._formatMessage(msg, standard_msg))

        if not np.all(arr1 == arr2):  # pragma: NO COVER
            standard_msg = NOT_EQ_TEMPLATE.format(arr1, arr2)
            self.fail(self._formatMessage(msg, standard_msg))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import numpy as np

from tests import utils


class Test_vector_close(unittest.TestCase):

    @staticmethod
    def _call_function_under_test(vec1, vec2, **kwargs):
        from bezier import _helpers

        return _helpers.vector_close(vec1, vec2, **kwargs)

    def test_identical(self):
        vec1 = np.array([0.5, 4.0])
        self.assertTrue(self._call_function_under_test(vec1, vec1))

    def test_far_apart(self):
        vec1 = np.array([0.0, 6.0])
        vec2 = np.array([1.0, -4.0])
        self.assertFalse(self._call_function_under_test(vec1, vec2))

    def test_close_but_different(self):
        vec1 = np.array([2.25, -3.5])
        vec2 = vec1 + np.array([-5.0, 12.0]) / 2.0**43
        self.assertTrue(self._call_function_under_test(vec1, vec2))

    def test_custom_epsilon(self):
        vec1 = np.array([3.0, 4.0])
        vec2 = np.array([2.0, 5.0])
        self.assertTrue(self._call_function_under_test(vec1, vec2, eps=0.5))
        self.assertFalse(self._call_function_under_test(vec1, vec2))

    def test_near_zero(self):
        vec1 = np.array([0.0, 0.0])
        vec2 = np.array([3.0, 4.0]) / 2.0**45
        self.assertTrue(self._call_function_under_test(vec1, vec2))

    def test_near_zero_fail(self):
        vec1 = np.array([1.0, 0.0]) / 2.0**20
        vec2 = np.array([0.0, 0.0])
        self.assertFalse(self._call_function_under_test(vec1, vec2))


class Test_in_interval(unittest.TestCase):

    @staticmethod
    def _call_function_under_test(value, start, end):
        from bezier import _helpers

        return _helpers.in_interval(value, start, end)

    def test_interior(self):
        self.assertTrue(self._call_function_under_test(
            1.5, 1.0, 2.0))

    def test_barely_inside(self):
        local_epsilon = np.spacing(1.0)  # pylint: disable=no-member
        self.assertTrue(self._call_function_under_test(
            1.0 + local_epsilon, 1.0, 2.0))

    def test_barely_outside(self):
        local_epsilon = np.spacing(1.0)  # pylint: disable=no-member
        self.assertFalse(self._call_function_under_test(
            1.0 - local_epsilon / 2.0, 1.0, 2.0))

    def test_outside(self):
        self.assertFalse(self._call_function_under_test(
            -1.0, 1.0, 2.0))


class Test_bbox(unittest.TestCase):

    @staticmethod
    def _call_function_under_test(nodes):
        from bezier import _helpers

        return _helpers.bbox(nodes)

    def test_it(self):
        nodes = np.array([
            [0.0, 5.0],
            [1.0, 3.0],
        ])
        left, right, bottom, top = self._call_function_under_test(nodes)
        self.assertEqual(left, 0.0)
        self.assertEqual(right, 1.0)
        self.assertEqual(bottom, 3.0)
        self.assertEqual(top, 5.0)


class Test_contains(unittest.TestCase):

    UNIT_SQUARE = np.array([
        [0.0, 0.0],
        [1.0, 1.0],
    ])

    @staticmethod
    def _call_function_under_test(nodes, x_val, y_val):
        from bezier import _helpers

        return _helpers.contains(nodes, x_val, y_val)

    def test_x_outside(self):
        self.assertFalse(
            self._call_function_under_test(self.UNIT_SQUARE, -1.0, 0.5))

    def test_y_outside(self):
        self.assertFalse(
            self._call_function_under_test(self.UNIT_SQUARE, 0.625, 4.25))

    def test_inside(self):
        self.assertTrue(
            self._call_function_under_test(self.UNIT_SQUARE, 0.25, 0.75))


class Test_contains_nd(unittest.TestCase):

    @staticmethod
    def _call_function_under_test(nodes, point):
        from bezier import _helpers

        return _helpers.contains_nd(nodes, point)

    def test_below(self):
        nodes = np.array([
            [0.0, 1.0],
            [0.5, 0.0],
            [1.0, 2.0],
        ])
        point = np.array([[-0.5, 1.0]])
        self.assertFalse(self._call_function_under_test(nodes, point))

    def test_above(self):
        nodes = np.array([
            [0.0, -4.0, 2.0],
            [-1.0, 1.0, 3.0],
        ])
        point = np.array([[-0.5, 2.0, 2.5]])
        self.assertFalse(self._call_function_under_test(nodes, point))

    def test_inside(self):
        nodes = np.array([
            [0.0, 1.0, 2.0, 3.0],
            [1.0, -2.0, -4.0, 1.0],
        ])
        point = np.array([[0.5, 0.0, 0.0, 2.0]])
        self.assertTrue(self._call_function_under_test(nodes, point))


class Test_cross_product(utils.NumPyTestCase):

    @staticmethod
    def _call_function_under_test(vec0, vec1):
        from bezier import _helpers

        return _helpers.cross_product(vec0, vec1)

    def test_it(self):
        vec0 = np.array([[1.0, 7.0]]) / 8.0
        vec1 = np.array([[-11.0, 24.0]]) / 32.0
        result = self._call_function_under_test(vec0, vec1)

        vec0_as_3d = np.hstack([vec0, [[0.0]]])
        vec1_as_3d = np.hstack([vec1, [[0.0]]])

        actual_cross = np.cross(vec0_as_3d, vec1_as_3d)
        expected = np.array([[0.0, 0.0, result]])
        self.assertEqual(actual_cross, expected)


class Test_n_bits_away(unittest.TestCase):

    @staticmethod
    def _call_function_under_test(value1, value2, **kwargs):
        from bezier import _helpers

        return _helpers.n_bits_away(value1, value2, **kwargs)

    def test_first_zero(self):
        self.assertTrue(
            self._call_function_under_test(0.0, 0.0))

    def test_second_zero(self):
        self.assertFalse(
            self._call_function_under_test(1.0, 0.0))

    def test_default_single_bit(self):
        value1 = 1.0
        value2 = 1.0 + 2.0**(-52)
        value3 = 1.0 - 2.0**(-53)

        self.assertTrue(
            self._call_function_under_test(value1, value1))
        self.assertTrue(
            self._call_function_under_test(value1, value2))
        self.assertTrue(
            self._call_function_under_test(value2, value1))
        self.assertTrue(
            self._call_function_under_test(value1, value3))
        self.assertTrue(
            self._call_function_under_test(value3, value1))

    def test_non_default_n(self):
        value1 = 1.5
        value2 = 1.5 + 2.0**(-43)

        self.assertFalse(
            self._call_function_under_test(value1, value2))
        self.assertTrue(
            self._call_function_under_test(value1, value2, num_bits=1000))

    def test_very_close(self):
        value1 = float.fromhex('0x1.fffffffffffd3p-3')
        value2 = float.fromhex('0x1.fffffffffffcep-3')

        self.assertFalse(
            self._call_function_under_test(value1, value2))
        self.assertFalse(
            self._call_function_under_test(value1, value2, num_bits=4))
        self.assertTrue(
            self._call_function_under_test(value1, value2, num_bits=5))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

import matplotlib.pyplot as plt
import numpy as np

import bezier
from bezier import _plot_helpers

import runtime_utils


CONFIG = runtime_utils.Config()


# F1 = sympy.Matrix([[s, t]])
SURFACE1 = bezier.Surface(np.array([
    [0.0, 0.0],
    [1.0, 0.0],
    [0.0, 1.0],
]))
# F2 = sympy.Matrix([[
#     (-t**2 + 2 * s + t) / 2, (s**2 + 2 * s * t - s + 2 * t) / 2]])
SURFACE2 = bezier.Surface(np.array([
    [0.0, 0.0],
    [0.5, -0.25],
    [1.0, 0.0],
    [0.25, 0.5],
    [0.75, 0.75],
    [0.0, 1.0],
]))
# F3 = sympy.Matrix([[
#     -(2 * s * t - 4 * s - t) / 4, (s**2 - s * t + 4 * t) / 4]])
SURFACE3 = bezier.Surface(np.array([
    [0.0, 0.0],
    [0.5, 0.0],
    [1.0, 0.25],
    [0.125, 0.5],
    [0.375, 0.375],
    [0.25, 1.0],
]))
# F4 = sympy.Matrix([[2 * (s + 2 * t) * (1 - t), 2 * t * (s + 1)]])
SURFACE4 = bezier.Surface(np.array([
    [0.0, 0.0],
    [1.0, 0.0],
    [2.0, 0.0],
    [2.0, 1.0],
    [2.0, 2.0],
    [0.0, 2.0],
]))

POINTS = np.array([
    [0.0, 0.0],
    [0.25, 0.25],
    [0.59375, 0.25],
    [0.265625, 0.73046875],
    [1.25, 1.25],
])


def make_plot(surface, point):
    if not CONFIG.running:
        return

    ax = surface.plot(64)
    ax.plot(point[:, 0], point[:, 1], color='black',
            marker='o', linestyle='None')

    ax.axis('scaled')
    _plot_helpers.add_plot_boundary(ax)

    if CONFIG.save_plot:
        CONFIG.save_fig()
    else:
        plt.title(CONFIG.current_test)
        plt.show()

    plt.close(ax.figure)


def check_point(surface, point_ind, expected_s, expected_t):
    point = POINTS[[point_ind], :]
    if expected_s is None and expected_t is None:
        assert surface.locate(point) is None
    else:
        s, t = surface.locate(point)

        CONFIG.assert_close(s, expected_s)
        CONFIG.assert_close(t, expected_t)

    make_plot(surface, point)


def test_surface1_and_point0():
    check_point(SURFACE1, 0, 0.0, 0.0)


def test_surface1_and_point1():
    check_point(SURFACE1, 1, 0.25, 0.25)


def test_surface2_and_point1():
    s, _ = runtime_utils.real_roots([4, -32, -56, -24, 5])
    _, t = runtime_utils.real_roots([4, 8, -16, 44, -11])
    check_point(SURFACE2, 1, s, t)


def test_surface2_and_point2():
    check_point(SURFACE2, 2, 0.5, 0.25)


def test_surface2_and_point4():
    check_point(SURFACE2, 4, None, None)


def test_surface3_and_point1():
    s, = runtime_utils.real_roots([2, -5, 15, -3])
    t, = runtime_utils.real_roots([14, -61, 74, -15])
    check_point(SURFACE3, 1, s, t)


def test_surface3_and_point3():
    check_point(SURFACE3, 3, 0.125, 0.75)


def test_surface4_and_point2():
    s, = runtime_utils.real_roots([64, 101, 34, -5])
    t, = runtime_utils.real_roots([128, -192, 91, -8])
    check_point(SURFACE4, 2, s, t)


def test_surface4_and_point4():
    check_point(SURFACE4, 4, 0.25, 0.5)


if __name__ == '__main__':
    CONFIG.run(globals())



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

"""Plotting utilities."""


import numpy as np

from bezier import _helpers


def add_plot_boundary(ax, padding=0.125):
    """Add a buffer of empty space around a plot boundary.

    .. note::

       This only uses ``line`` data from the axis. It **could**
       use ``patch`` data, but doesn't at this time.

    Args:
        ax (matplotlib.artist.Artist): A matplotlib axis.
        padding (Optional[float]): Amount (as a fraction of width and height)
            of padding to add around data. Defaults to ``0.125``.
    """
    nodes = np.vstack([line.get_xydata() for line in ax.lines])
    left, right, bottom, top = _helpers.bbox(nodes)
    center_x = 0.5 * (right + left)
    delta_x = right - left
    center_y = 0.5 * (top + bottom)
    delta_y = top - bottom

    multiplier = (1.0 + padding) * 0.5
    ax.set_xlim(center_x - multiplier * delta_x,
                center_x + multiplier * delta_x)
    ax.set_ylim(center_y - multiplier * delta_y,
                center_y + multiplier * delta_y)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import mock
import numpy as np

from tests import utils


# pylint: disable=too-many-arguments
def check_intersection(test_case, intersection, expected,
                       curve1, curve2, s_val, t_val):
    from bezier import _intersection_helpers

    test_case.assertIsInstance(
        intersection, _intersection_helpers.Intersection)
    test_case.assertEqual(intersection.point, expected)
    test_case.assertIs(intersection.left, curve1)
    test_case.assertEqual(intersection.s, s_val)
    test_case.assertIs(intersection.right, curve2)
    test_case.assertEqual(intersection.t, t_val)
# pylint: enable=too-many-arguments


class Test__check_close(utils.NumPyTestCase):

    @staticmethod
    def _call_function_under_test(s, curve1, t, curve2):
        from bezier import _intersection_helpers

        return _intersection_helpers._check_close(
            s, curve1, t, curve2)

    def test_success(self):
        import bezier

        nodes = np.array([
            [0.0, 0.0],
            [0.5, 1.0],
            [1.0, 0.0],
        ])
        curve = bezier.Curve(nodes)
        s_val = 0.5
        wiggle = 2.0**(-50)
        result = self._call_function_under_test(
            s_val, curve, s_val + wiggle, curve)

        expected = np.array([0.5, 0.5])
        self.assertEqual(result, expected)

    def test_failure(self):
        import bezier

        nodes = np.array([
            [0.0, 0.0],
            [1.0, 1.0],
        ])
        curve = bezier.Curve(nodes)
        # The nodes of thise curve are far away.
        with self.assertRaises(ValueError):
            self._call_function_under_test(0.0, curve, 1.0, curve)


class Test__check_parameters(unittest.TestCase):

    @staticmethod
    def _call_function_under_test(s, t):
        from bezier import _intersection_helpers

        return _intersection_helpers._check_parameters(s, t)

    def test_at_endpoint(self):
        # Really just making sure the function doesn't raise.
        result = self._call_function_under_test(0.0, 0.5)
        self.assertEqual(result, (0.0, 0.5))
        result = self._call_function_under_test(0.5, 1.0)
        self.assertEqual(result, (0.5, 1.0))

    def test_near_endpoint(self):
        with self.assertRaises(ValueError):
            self._call_function_under_test(0.75, 1.0 + 2.0**(-20))

    def test_s_outside(self):
        with self.assertRaises(ValueError):
            self._call_function_under_test(-0.25, 0.5)

    def test_t_outside(self):
        with self.assertRaises(ValueError):
            self._call_function_under_test(0.25, 1.5)

    def test_valid(self):
        # Really just making sure the function doesn't raise.
        result = self._call_function_under_test(0.25, 0.5)
        self.assertEqual(result, (0.25, 0.5))

    def test_wiggle_below(self):
        value = -2.0**(-60)
        result = self._call_function_under_test(value, 0.25)
        self.assertEqual(result, (0.0, 0.25))

    def test_wiggle_above(self):
        value = 1 + 2.0**(-52)
        result = self._call_function_under_test(0.875, value)
        self.assertEqual(result, (0.875, 1.0))


class Test_bbox_intersect(unittest.TestCase):

    UNIT_SQUARE = np.array([
        [0.0, 0.0],
        [1.0, 0.0],
        [1.0, 1.0],
        [0.0, 1.0],
    ])

    @staticmethod
    def _call_function_under_test(nodes1, nodes2):
        from bezier import _intersection_helpers

        return _intersection_helpers.bbox_intersect(nodes1, nodes2)

    def test_intersect(self):
        from bezier import _intersection_helpers

        nodes = self.UNIT_SQUARE + np.array([[0.5, 0.5]])
        result = self._call_function_under_test(self.UNIT_SQUARE, nodes)
        self.assertIs(
            result, _intersection_helpers.BoxIntersectionType.intersection)

    def test_far_apart(self):
        from bezier import _intersection_helpers

        nodes = self.UNIT_SQUARE + np.array([[100.0, 100.0]])
        result = self._call_function_under_test(self.UNIT_SQUARE, nodes)
        self.assertIs(
            result, _intersection_helpers.BoxIntersectionType.disjoint)

    def test_disjoint_but_aligned(self):
        from bezier import _intersection_helpers

        nodes = self.UNIT_SQUARE + np.array([[1.0, 2.0]])
        result = self._call_function_under_test(self.UNIT_SQUARE, nodes)
        self.assertIs(
            result, _intersection_helpers.BoxIntersectionType.disjoint)

    def test_tangent(self):
        from bezier import _intersection_helpers

        nodes = self.UNIT_SQUARE + np.array([[1.0, 0.0]])
        result = self._call_function_under_test(self.UNIT_SQUARE, nodes)
        self.assertIs(
            result, _intersection_helpers.BoxIntersectionType.tangent)

    def test_almost_tangent(self):
        from bezier import _intersection_helpers

        x_val = 1.0 + np.spacing(1.0)  # pylint: disable=no-member
        nodes = self.UNIT_SQUARE + np.array([[x_val, 0.0]])
        result = self._call_function_under_test(self.UNIT_SQUARE, nodes)
        self.assertIs(
            result, _intersection_helpers.BoxIntersectionType.disjoint)


class Test_linearization_error(unittest.TestCase):

    @staticmethod
    def _call_function_under_test(curve):
        from bezier import _intersection_helpers

        return _intersection_helpers.linearization_error(curve)

    def test_linear(self):
        import bezier

        nodes = np.array([
            [0.0, 0.0],
            [1.0, 2.0],
        ])
        curve = bezier.Curve(nodes)
        error_val = self._call_function_under_test(curve)
        self.assertEqual(error_val, 0.0)

    def test_degree_elevated_linear(self):
        import bezier

        nodes = np.array([
            [0.0, 0.0],
            [0.5, 1.0],
            [1.0, 2.0],
        ])
        curve = bezier.Curve(nodes)
        error_val = self._call_function_under_test(curve)
        self.assertEqual(error_val, 0.0)

        nodes = np.array([
            [0.0, 0.0],
            [0.25, 0.5],
            [0.5, 1.0],
            [0.75, 1.5],
            [1.0, 2.0],
        ])
        curve = bezier.Curve(nodes)
        error_val = self._call_function_under_test(curve)
        self.assertEqual(error_val, 0.0)

    def test_hidden_linear(self):
        import bezier

        # NOTE: This is the line 3 y = 4 x, but with the parameterization
        #       x(s) = 3 s (4 - 3 s).
        nodes = np.array([
            [0.0, 0.0],
            [6.0, 8.0],
            [3.0, 4.0],
        ])
        curve = bezier.Curve(nodes)
        error_val = self._call_function_under_test(curve)
        # D^2 v = [-9, -12]
        expected = 0.125 * 2 * 1 * 15.0
        self.assertEqual(error_val, expected)

    def test_quadratic(self):
        import bezier

        nodes = np.array([
            [0.0, 0.0],
            [1.0, 1.0],
            [5.0, 6.0],
        ])
        # NOTE: This is hand picked so that
        #             d Nodes = [1, 1], [4, 5]
        #           d^2 Nodes = [3, 4]
        #       so that sqrt(3^2 + 4^2) = 5.0
        curve = bezier.Curve(nodes)
        error_val = self._call_function_under_test(curve)
        expected = 0.125 * 2 * 1 * 5.0
        self.assertEqual(error_val, expected)

        # For a degree two curve, the 2nd derivative is constant
        # so by subdividing, our error should drop by a factor
        # of (1/2)^2 = 4.
        left, right = curve.subdivide()
        error_left = self._call_function_under_test(left)
        error_right = self._call_function_under_test(right)
        self.assertEqual(error_left, 0.25 * expected)
        self.assertEqual(error_right, 0.25 * expected)

    def test_hidden_quadratic(self):
        import bezier

        # NOTE: This is the line y = 1 + x^2 / 4, but with the
        #       parameterization x(s) = (3 s - 1)^2.
        nodes = np.array([
            [1.0, 1.25],
            [-0.5, 0.5],
            [-0.5, 2.0],
            [1.0, -1.0],
            [4.0, 5.0],
        ])
        curve = bezier.Curve(nodes)
        error_val = self._call_function_under_test(curve)
        # D^2 v = [1.5, 2.25], [1.5, -4.5], [1.5, 9]
        expected = 0.125 * 4 * 3 * np.sqrt(1.5**2 + 9.0**2)
        self.assertEqual(error_val, expected)

    def test_cubic(self):
        import bezier

        nodes = np.array([
            [0.0, 0.0],
            [1.0, 1.0],
            [5.0, 6.0],
            [6.0, 7.0],
        ])
        # NOTE: This is hand picked so that
        #             d Nodes = [1, 1], [4, 5], [1, 1]
        #           d^2 Nodes = [3, 4], [-3, -4]
        #       so that sqrt(3^2 + 4^2) = 5.0
        curve = bezier.Curve(nodes)
        error_val = self._call_function_under_test(curve)
        expected = 0.125 * 3 * 2 * 5.0
        self.assertEqual(error_val, expected)

    def test_quartic(self):
        import bezier

        nodes = np.array([
            [0.0, 0.0],
            [1.0, 1.0],
            [5.0, 6.0],
            [6.0, 7.0],
            [4.0, 7.0],
        ])
        # NOTE: This is hand picked so that
        #             d Nodes = [1, 1], [4, 5], [1, 1], [-2, 0]
        #           d^2 Nodes = [3, 4], [-3, -4], [-3, -1]
        #       so that sqrt(3^2 + 4^2) = 5.0
        curve = bezier.Curve(nodes)
        error_val = self._call_function_under_test(curve)
        expected = 0.125 * 4 * 3 * 5.0
        self.assertEqual(error_val, expected)

    def test_degree_weights_on_the_fly(self):
        import bezier

        nodes = np.array([
            [0.0, 0.0],
            [1.0, 1.0],
            [7.0, 3.0],
            [11.0, 8.0],
            [15.0, 1.0],
            [16.0, -3.0],
        ])
        # NOTE: This is hand picked so that
        #             d Nodes = [1, 1], [6, 2], [4, 5], [4, -7], [1, -4]
        #           d^2 Nodes = [5, 1], [-2, 3], [0, -12], [-3, 3]
        #       so that sqrt(5^2 + 12^2) = 13.0
        curve = bezier.Curve(nodes)
        error_val = self._call_function_under_test(curve)
        expected = 0.125 * 5 * 4 * 13.0
        self.assertEqual(error_val, expected)


class Test_newton_refine(utils.NumPyTestCase):

    @staticmethod
    def _call_function_under_test(s, curve1, t, curve2):
        from bezier import _intersection_helpers

        return _intersection_helpers.newton_refine(s, curve1, t, curve2)

    def test_linear(self):
        import bezier

        nodes1 = np.array([
            [0.0, 0.0],
            [1.0, 1.0],
        ])
        nodes2 = np.array([
            [1.0, 0.0],
            [0.0, 3.0],
        ])
        curve1 = bezier.Curve(nodes1)
        curve2 = bezier.Curve(nodes2)

        known_s = 0.75
        known_t = 0.25
        self.assertEqual(curve1.evaluate(known_s),
                         curve2.evaluate(known_t))

        wrong_s = known_s - 0.125
        wrong_t = known_t + 0.125
        # NOTE: By construction, the Jacobian matrix will be
        #           [1, 1], [1, -3]
        #       which has determinant -4.0, hence there will
        #       be no round-off when solving.
        new_s, new_t = self._call_function_under_test(
            wrong_s, curve1, wrong_t, curve2)

        # Newton's method is exact on linear problems so will
        # always converge after one step.
        self.assertEqual(new_s, known_s)
        self.assertEqual(new_t, known_t)

    @staticmethod
    def _get_quadratics():
        import bezier

        nodes1 = np.array([
            [0.0, 0.0],
            [0.5, 1.0],
            [1.0, 0.0],
        ])
        nodes2 = np.array([
            [1.0, 0.75],
            [0.5, -0.25],
            [0.0, 0.75],
        ])
        curve1 = bezier.Curve(nodes1)
        curve2 = bezier.Curve(nodes2)
        return curve1, curve2

    def test_mixed_degree(self):
        import bezier

        curve1, _ = self._get_quadratics()
        nodes2 = np.array([
            [1.0, 0.0],
            [0.0, 1.0],
        ])
        curve2 = bezier.Curve(nodes2)

        known_s = 0.5
        known_t = 0.5
        self.assertEqual(curve1.evaluate(known_s),
                         curve2.evaluate(known_t))

        wrong_s = 0.25
        wrong_t = 0.25
        # NOTE: By construction, the Jacobian matrix will be
        #           [1, 1], [1, -1]
        #       which has determinant -2.0, hence there will
        #       be no round-off when solving.
        new_s, new_t = self._call_function_under_test(
            wrong_s, curve1, wrong_t, curve2)

        self.assertEqual(new_s, 0.4375)
        self.assertEqual(new_t, 0.5625)

        # Make sure we have gotten closer to correct.
        self.assertLess(abs(known_s - new_s), abs(known_s - wrong_s))
        self.assertLess(abs(known_t - new_t), abs(known_t - wrong_t))

    def test_early_exit(self):
        curve1, curve2 = self._get_quadratics()

        known_s = 0.25
        known_t = 0.75

        self.assertEqual(curve1.evaluate(known_s),
                         curve2.evaluate(known_t))

        new_s, new_t = self._call_function_under_test(
            known_s, curve1, known_t, curve2)

        self.assertEqual(new_s, known_s)
        self.assertEqual(new_t, known_t)

    def test_quadratic(self):
        curve1, curve2 = self._get_quadratics()

        known_s = 0.25
        known_t = 0.75
        self.assertEqual(curve1.evaluate(known_s),
                         curve2.evaluate(known_t))

        wrong_s = known_s + 0.0625  # 1/16
        wrong_t = known_t + 0.0625  # 1/16
        # NOTE: By construction, the Jacobian matrix will be
        #           [1, 3/4], [1, -5/4]
        #       which has determinant -2.0, hence there will
        #       be no round-off when solving.
        new_s, new_t = self._call_function_under_test(
            wrong_s, curve1, wrong_t, curve2)

        self.assertEqual(new_s, 0.2421875)
        self.assertEqual(new_t, 0.7578125)

        # Make sure we have gotten closer to correct.
        self.assertLess(abs(known_s - new_s), abs(known_s - wrong_s))
        self.assertLess(abs(known_t - new_t), abs(known_t - wrong_t))

    def test_convergence(self):
        import six

        import bezier

        nodes1 = np.array([
            [0.0, 0.0],
            [0.25, 1.0],
            [0.5, -0.75],
            [0.75, 1.0],
            [1.0, 0.0],
        ])
        curve1 = bezier.Curve(nodes1)
        # Vertical line forces a unique solution.
        nodes2 = np.array([
            [0.5, 0.0],
            [0.5, 1.0],
        ])
        curve2 = bezier.Curve(nodes2)

        num_guess = 4
        parameters = np.zeros((num_guess, 2))
        # NOTE: This means our "first" guess is (s, t) = (0, 0).
        for guess in six.moves.xrange(1, num_guess):
            prev_s, prev_t = parameters[guess - 1, :]
            parameters[guess, :] = self._call_function_under_test(
                prev_s, curve1, prev_t, curve2)

        expected = np.array([
            [0.0, 0.0],
            [0.5, 2.0],
            [0.5, 0.21875],
            [0.5, 0.21875],
        ])
        self.assertEqual(parameters, expected)
        # Make sure that we've actually converged.
        exact_s, exact_t = parameters[-1, :]
        self.assertEqual(curve1.evaluate(exact_s),
                         curve2.evaluate(exact_t))


class Test_segment_intersection(unittest.TestCase):

    @staticmethod
    def _call_function_under_test(start0, end0, start1, end1, **kwargs):
        from bezier import _intersection_helpers

        return _intersection_helpers.segment_intersection(
            start0, end0, start1, end1, **kwargs)

    def _helper(self, intersection, s_val, direction0,
                t_val, direction1, **kwargs):
        start0 = intersection + s_val * direction0
        end0 = intersection + (s_val - 1.0) * direction0
        start1 = intersection + t_val * direction1
        end1 = intersection + (t_val - 1.0) * direction1

        return self._call_function_under_test(
            start0, end0, start1, end1, **kwargs)

    def test_success(self):
        intersection = np.array([[1.0, 2.0]])
        s_val = 0.25
        t_val = 0.625
        direction0 = np.array([[3.0, 0.5]])
        direction1 = np.array([[-2.0, 1.0]])
        # D0 x D1 == 4.0, so there will be no round-off in answer.
        computed_s, computed_t, success = self._helper(
            intersection, s_val, direction0, t_val, direction1)

        self.assertEqual(computed_s, s_val)
        self.assertEqual(computed_t, t_val)
        self.assertTrue(success)

    def test_parallel(self):
        intersection = np.array([[0.0, 0.0]])
        s_val = 0.5
        t_val = 0.5
        direction0 = np.array([[0.0, 1.0]])
        direction1 = np.array([[0.0, 2.0]])
        computed_s, computed_t, success = self._helper(
            intersection, s_val,
            direction0, t_val, direction1)

        self.assertIsNone(computed_s)
        self.assertIsNone(computed_t)
        self.assertFalse(success)


class Test_parallel_different(unittest.TestCase):

    @staticmethod
    def _call_function_under_test(start0, end0, start1, end1):
        from bezier import _intersection_helpers

        return _intersection_helpers.parallel_different(
            start0, end0, start1, end1)

    def test_same_line_no_overlap(self):
        start0 = np.array([[0.0, 0.0]])
        end0 = np.array([[3.0, 4.0]])
        start1 = np.array([[6.0, 8.0]])
        end1 = np.array([[9.0, 12.0]])
        self.assertTrue(
            self._call_function_under_test(start0, end0, start1, end1))

    def test_same_line_overlap_at_start(self):
        start0 = np.array([[6.0, -3.0]])
        end0 = np.array([[-7.0, 1.0]])
        start1 = np.array([[1.125, -1.5]])
        end1 = np.array([[-5.375, 0.5]])
        self.assertFalse(
            self._call_function_under_test(start0, end0, start1, end1))

    def test_same_line_overlap_at_end(self):
        start0 = np.array([[1.0, 2.0]])
        end0 = np.array([[3.0, 5.0]])
        start1 = np.array([[-0.5, -0.25]])
        end1 = np.array([[2.0, 3.5]])
        self.assertFalse(
            self._call_function_under_test(start0, end0, start1, end1))

    def test_same_line_contained(self):
        start0 = np.array([[-9.0, 0.0]])
        end0 = np.array([[4.0, 5.0]])
        start1 = np.array([[23.5, 12.5]])
        end1 = np.array([[-25.25, -6.25]])
        self.assertFalse(
            self._call_function_under_test(start0, end0, start1, end1))

    def test_different_line(self):
        start0 = np.array([[3.0, 2.0]])
        end0 = np.array([[3.0, 0.75]])
        start1 = np.array([[0.0, 0.0]])
        end1 = np.array([[0.0, 2.0]])
        self.assertTrue(
            self._call_function_under_test(start0, end0, start1, end1))


class Test_from_linearized(utils.NumPyTestCase):

    @staticmethod
    def _call_function_under_test(left, right, intersections):
        from bezier import _intersection_helpers

        return _intersection_helpers.from_linearized(
            left, right, intersections)

    def test_it(self):
        import bezier
        from bezier import _intersection_helpers

        nodes1 = np.array([
            [0.0, 0.0],
            [0.5, 1.0],
            [1.0, 1.0],
        ])
        curve1 = bezier.Curve(nodes1)
        # NOTE: This curve isn't close to linear, but that's OK.
        lin1 = _intersection_helpers.Linearization(curve1)

        nodes2 = np.array([
            [0.0, 1.0],
            [0.5, 1.0],
            [1.0, 0.0],
        ])
        curve2 = bezier.Curve(nodes2)
        # NOTE: This curve isn't close to linear, but that's OK.
        lin2 = _intersection_helpers.Linearization(curve2)

        intersections = []
        self.assertIsNone(
            self._call_function_under_test(lin1, lin2, intersections))
        self.assertEqual(len(intersections), 1)
        intersection = intersections[0]
        expected = curve1.evaluate(0.5)
        check_intersection(self, intersection, expected,
                           curve1, curve2, 0.5, 0.5)

    def test_no_intersection(self):
        import bezier
        from bezier import _intersection_helpers
        # The bounding boxes intersect but the lines do not.

        nodes1 = np.array([
            [0.0, 0.0],
            [1.0, 1.0],
        ])
        curve1 = bezier.Curve(nodes1)
        lin1 = _intersection_helpers.Linearization(curve1)

        nodes2 = np.array([
            [1.75, -0.75],
            [0.75, 0.25],
        ])
        curve2 = bezier.Curve(nodes2)
        lin2 = _intersection_helpers.Linearization(curve2)

        intersections = []
        self.assertIsNone(
            self._call_function_under_test(lin1, lin2, intersections))
        self.assertEqual(intersections, [])

    def test_parallel_intersection(self):
        import bezier
        from bezier import _intersection_helpers

        nodes1 = np.array([
            [0.0, 0.0],
            [1.0, 1.0],
        ])
        curve1 = bezier.Curve(nodes1)
        lin1 = _intersection_helpers.Linearization(curve1)

        nodes2 = np.array([
            [0.0, 1.0],
            [1.0, 2.0],
        ])
        curve2 = bezier.Curve(nodes2)
        lin2 = _intersection_helpers.Linearization(curve2)

        intersections = []
        self.assertIsNone(
            self._call_function_under_test(lin1, lin2, intersections))
        self.assertEqual(intersections, [])

    def test_same_line_intersection(self):
        import bezier
        from bezier import _intersection_helpers

        nodes1 = np.array([
            [0.0, 0.0],
            [1.0, 1.0],
        ])
        curve1 = bezier.Curve(nodes1)
        lin1 = _intersection_helpers.Linearization(curve1)

        nodes2 = np.array([
            [0.5, 0.5],
            [3.0, 3.0],
        ])
        curve2 = bezier.Curve(nodes2)
        lin2 = _intersection_helpers.Linearization(curve2)

        with self.assertRaises(NotImplementedError):
            self._call_function_under_test(lin1, lin2, [])

    def test_parallel_non_degree_one(self):
        import bezier
        from bezier import _intersection_helpers

        nodes1 = np.array([
            [0.0, 0.0],
            [1.0, 1.0],
        ])
        curve1 = bezier.Curve(nodes1)
        lin1 = _intersection_helpers.Linearization(curve1)

        nodes2 = np.array([
            [2.0, 2.0],
            [2.5009765625, 2.5009765625],
            [3.0, 3.0],
        ])
        curve2 = bezier.Curve(nodes2)
        lin2 = _intersection_helpers.Linearization(curve2)

        with self.assertRaises(NotImplementedError):
            self._call_function_under_test(lin1, lin2, [])


class Test__add_intersection(unittest.TestCase):

    @staticmethod
    def _call_function_under_test(intersection, intersections):
        from bezier import _intersection_helpers

        return _intersection_helpers._add_intersection(
            intersection, intersections)

    def test_new(self):
        from bezier import _intersection_helpers

        intersection1 = _intersection_helpers.Intersection(
            mock.sentinel.left, 0.5, mock.sentinel.right, 0.5)
        intersection2 = _intersection_helpers.Intersection(
            mock.sentinel.left, 0.75, mock.sentinel.right, 0.25)
        intersections = [intersection1]
        self.assertIsNone(
            self._call_function_under_test(intersection2, intersections))

        self.assertEqual(intersections, [intersection1, intersection2])

    def test_existing(self):
        from bezier import _intersection_helpers

        intersection1 = _intersection_helpers.Intersection(
            mock.sentinel.left, 0.0, mock.sentinel.right, 1.0)
        intersection2 = _intersection_helpers.Intersection(
            mock.sentinel.left, 0.0, mock.sentinel.right, 1.0)
        intersections = [intersection1]
        self.assertIsNone(
            self._call_function_under_test(intersection2, intersections))

        self.assertEqual(len(intersections), 1)
        self.assertIs(intersections[0], intersection1)
        self.assertIsNot(intersections[0], intersection2)


class Test__tangent_bbox_intersection(utils.NumPyTestCase):

    @staticmethod
    def _call_function_under_test(left, right, intersections):
        from bezier import _intersection_helpers

        return _intersection_helpers._tangent_bbox_intersection(
            left, right, intersections)

    def test_it(self):
        import bezier

        nodes1 = np.array([
            [0.0, 0.0],
            [1.0, 2.0],
            [2.0, 0.0],
        ])
        curve1 = bezier.Curve(nodes1)
        nodes2 = np.array([
            [2.0, 0.0],
            [3.0, 2.0],
            [4.0, 0.0],
        ])
        curve2 = bezier.Curve(nodes2)

        intersections = []
        self.assertIsNone(
            self._call_function_under_test(curve1, curve2, intersections))
        self.assertEqual(len(intersections), 1)
        intersection = intersections[0]
        expected = curve1.evaluate(1.0)
        check_intersection(self, intersection, expected,
                           curve1, curve2, 1.0, 0.0)


class Test_intersect_one_round(utils.NumPyTestCase):

    # NOTE: NODES1 is a specialization of [0, 0], [1/2, 1], [1, 1]
    #       onto the interval [1/4, 1].
    NODES1 = np.array([
        [0.25, 0.4375],
        [0.625, 1.0],
        [1.0, 1.0],
    ])
    # NOTE: NODES2 is a specialization of [0, 1], [1/2, 1], [1, 0]
    #       onto the interval [0, 3/4].
    NODES2 = np.array([
        [0.0, 1.0],
        [0.375, 1.0],
        [0.75, 0.4375],
    ])
    LINE1 = np.array([
        [0.0, 0.0],
        [1.0, 1.0],
    ])
    LINE2 = np.array([
        [0.0, 1.0],
        [1.0, 0.0],
    ])

    @staticmethod
    def _call_function_under_test(candidates, intersections):
        from bezier import _intersection_helpers

        return _intersection_helpers.intersect_one_round(
            candidates, intersections)

    def test_simple(self):
        import bezier

        curve1 = bezier.Curve(self.NODES1)
        curve2 = bezier.Curve(self.NODES2)
        candidates = [(curve1, curve2)]
        accepted = self._call_function_under_test(
            candidates, [])

        self.assertEqual(accepted, [(curve1, curve2)])

    def test_left_linearized(self):
        import bezier
        from bezier import _intersection_helpers

        curve1 = bezier.Curve(self.LINE1)
        lin1 = _intersection_helpers.Linearization(curve1)
        curve2 = bezier.Curve(self.LINE2)

        intersections = []
        accepted = self._call_function_under_test(
            [(lin1, curve2)], intersections)

        self.assertEqual(intersections, [])
        self.assertEqual(accepted, [(lin1, curve2)])

    def test_right_linearized(self):
        import bezier
        from bezier import _intersection_helpers

        curve1 = bezier.Curve(self.LINE1)
        curve2 = bezier.Curve(self.LINE2)
        lin2 = _intersection_helpers.Linearization(curve2)

        intersections = []
        accepted = self._call_function_under_test(
            [(curve1, lin2)], intersections)

        self.assertEqual(intersections, [])
        self.assertEqual(accepted, [(curve1, lin2)])

    def test_both_linearized(self):
        import bezier
        from bezier import _intersection_helpers

        curve1 = bezier.Curve(self.LINE1)
        lin1 = _intersection_helpers.Linearization(curve1)
        curve2 = bezier.Curve(self.LINE2)
        lin2 = _intersection_helpers.Linearization(curve2)

        intersections = []
        accepted = self._call_function_under_test(
            [(lin1, lin2)], intersections)
        self.assertEqual(accepted, [])
        self.assertEqual(len(intersections), 1)
        intersection = intersections[0]
        expected = np.array([0.5, 0.5])
        check_intersection(self, intersection, expected,
                           curve1, curve2, 0.5, 0.5)


class Test__next_candidates(unittest.TestCase):

    @staticmethod
    def _call_function_under_test(left, right):
        from bezier import _intersection_helpers

        return _intersection_helpers._next_candidates(left, right)

    def test_it(self):
        import types
        import bezier
        from bezier import _intersection_helpers

        nodes = np.array([
            [0.0, 0.0],
            [1.0, 1.0],
        ])
        curve = bezier.Curve(nodes)
        lin = _intersection_helpers.Linearization(curve)

        result = self._call_function_under_test(lin, lin)

        self.assertIsInstance(result, types.GeneratorType)
        pairs = list(result)
        self.assertEqual(pairs, [(lin, lin)])


class Test_all_intersections(utils.NumPyTestCase):

    @staticmethod
    def _call_function_under_test(candidates):
        from bezier import _intersection_helpers

        return _intersection_helpers.all_intersections(candidates)

    def test_failure(self):
        patch = mock.patch(
            'bezier._intersection_helpers._MAX_INTERSECT_SUBDIVISIONS',
            new=-1)
        with patch:
            with self.assertRaises(ValueError):
                self._call_function_under_test([])

    def test_no_intersections(self):
        intersections = self._call_function_under_test([])
        self.assertEqual(intersections, [])

    def test_tangent(self):
        import itertools
        import bezier

        nodes = np.array([
            [0.0, 0.0],
            [0.5, 1.0],
            [1.0, 0.0],
        ])
        curve = bezier.Curve(nodes)

        # Start with two pieces.
        pieces = itertools.chain(*[curve.subdivide()])
        # Double the number of pieces 3 times.
        for unused_exponent in (2, 3, 4):
            pieces = itertools.chain(*[
                piece.subdivide() for piece in pieces])

        # Match each piece with itself.
        candidates = [(piece, piece) for piece in pieces]
        # But we need **more** than 16.
        candidates.append((curve, curve))

        with self.assertRaises(NotImplementedError):
            self._call_function_under_test(candidates)

    def test_success(self):
        import bezier

        # NOTE: ``nodes1`` is a specialization of [0, 0], [1/2, 1], [1, 1]
        #       onto the interval [1/4, 1] and ``nodes`` is a specialization
        #       of [0, 1], [1/2, 1], [1, 0] onto the interval [0, 3/4].
        #       We expect them to intersect at s = 1/3, t = 2/3, which is
        #       the point [1/2, 3/4].
        nodes1 = np.array([
            [0.25, 0.4375],
            [0.625, 1.0],
            [1.0, 1.0],
        ])
        curve1 = bezier.Curve(nodes1)

        nodes2 = np.array([
            [0.0, 1.0],
            [0.375, 1.0],
            [0.75, 0.4375],
        ])
        curve2 = bezier.Curve(nodes2)

        candidates = [(curve1, curve2)]
        intersections = self._call_function_under_test(candidates)

        self.assertEqual(len(intersections), 1)
        intersection = intersections[0]
        expected = np.array([0.5, 0.75])

        s_val = 1.0 / 3.0
        # Due to round-off, the answer is wrong by a tiny wiggle.
        s_val += np.spacing(s_val)  # pylint: disable=no-member
        t_val = 2.0 / 3.0
        check_intersection(self, intersection, expected,
                           curve1, curve2, s_val, t_val)


class TestLinearization(utils.NumPyTestCase):

    NODES = np.array([
        [0.0, 0.0],
        [1.0, 1.0],
        [5.0, 6.0],
    ])

    @staticmethod
    def _get_target_class():
        from bezier import _intersection_helpers

        return _intersection_helpers.Linearization

    def _make_one(self, *args, **kwargs):
        klass = self._get_target_class()
        return klass(*args, **kwargs)

    def test_constructor(self):
        linearization = self._make_one(mock.sentinel.curve)
        self.assertIs(linearization._curve, mock.sentinel.curve)
        self.assertIsNone(linearization._error)

    def test_constructor_with_error(self):
        error = 0.125
        linearization = self._make_one(mock.sentinel.curve, error=error)
        self.assertIs(linearization._curve, mock.sentinel.curve)
        self.assertEqual(linearization._error, error)

    def test_subdivide(self):
        linearization = self._make_one(None)
        self.assertEqual(linearization.subdivide(), (linearization,))

    def test_curve_property(self):
        linearization = self._make_one(mock.sentinel.curve)
        self.assertIs(linearization.curve, mock.sentinel.curve)

    def test_error_property(self):
        error = 0.0625
        linearization = self._make_one(None, error=error)
        self.assertEqual(linearization.error, error)

    def test_error_property_on_the_fly(self):
        linearization = self._make_one(mock.sentinel.curve)

        error = 0.09375
        patch = mock.patch(
            'bezier._intersection_helpers.linearization_error',
            return_value=error)
        with patch as mocked:
            self.assertEqual(linearization.error, error)
            mocked.assert_called_once_with(mock.sentinel.curve)

    def test_start_node_property(self):
        import bezier

        curve = bezier.Curve(self.NODES, _copy=False)
        linearization = self._make_one(curve)
        expected = self.NODES[[0], :]
        self.assertEqual(linearization.start_node, expected)

    def test_end_node_property(self):
        import bezier

        curve = bezier.Curve(self.NODES, _copy=False)
        linearization = self._make_one(curve)
        expected = self.NODES[[2], :]
        self.assertEqual(linearization.end_node, expected)

    def test_from_shape_factory_not_close_enough(self):
        import bezier

        curve = bezier.Curve(self.NODES, _copy=False)
        klass = self._get_target_class()
        new_shape = klass.from_shape(curve)
        self.assertIs(new_shape, curve)

    def test_from_shape_factory_close_enough(self):
        import bezier

        scale_factor = 2.0**(-27)
        nodes = self.NODES * scale_factor
        curve = bezier.Curve(nodes, _copy=False)
        klass = self._get_target_class()
        new_shape = klass.from_shape(curve)

        self.assertIsInstance(new_shape, klass)
        self.assertIs(new_shape._curve, curve)
        # NODES has constant second derivative equal to 2 * [3.0, 4.0].
        expected_error = 0.125 * 2 * 1 * 5.0 * scale_factor
        self.assertEqual(new_shape.error, expected_error)

    def test_from_shape_factory_no_error(self):
        import bezier

        nodes = np.array([
            [0.0, 0.0],
            [1.0, 1.0],
        ])
        curve = bezier.Curve(nodes, _copy=False)
        klass = self._get_target_class()
        new_shape = klass.from_shape(curve)
        self.assertIsInstance(new_shape, klass)
        self.assertIs(new_shape._curve, curve)
        # ``nodes`` is linear, so error is 0.0.
        self.assertEqual(new_shape.error, 0.0)

    def test_from_shape_factory_already_linearized(self):
        error = 0.078125
        linearization = self._make_one(None, error=error)

        klass = self._get_target_class()
        new_shape = klass.from_shape(linearization)
        self.assertIs(new_shape, linearization)
        self.assertEqual(new_shape.error, error)


class TestIntersection(unittest.TestCase):

    @staticmethod
    def _get_target_class():
        from bezier import _intersection_helpers

        return _intersection_helpers.Intersection

    def _make_one(self, *args, **kwargs):
        klass = self._get_target_class()
        return klass(*args, **kwargs)

    def _constructor_helper(self, **kwargs):
        left = mock.sentinel.left
        s_val = 0.25
        right = mock.sentinel.right
        t_val = 0.75

        intersection = self._make_one(
            left, s_val, right, t_val, **kwargs)

        self.assertIs(intersection._left, left)
        self.assertEqual(intersection._s_val, s_val)
        self.assertIs(intersection._right, right)
        self.assertEqual(intersection._t_val, t_val)
        self.assertIsNone(intersection._interior_curve)
        return intersection

    def test_constructor(self):
        intersection = self._constructor_helper()
        self.assertIsNone(intersection._point)

    def test_constructor_with_point(self):
        intersection = self._constructor_helper(point=mock.sentinel.point)
        self.assertIs(intersection._point, mock.sentinel.point)

    def test_s_property(self):
        s = 0.375
        intersection = self._make_one(None, s, None, None)
        self.assertEqual(intersection.s, s)

    def test_t_property(self):
        t = 0.9990234375
        intersection = self._make_one(None, None, None, t)
        self.assertEqual(intersection.t, t)

    def test_left_property(self):
        intersection = self._make_one(
            mock.sentinel.left, None, None, None)
        self.assertIs(intersection.left, mock.sentinel.left)

    def test_right_property(self):
        intersection = self._make_one(
            None, None, mock.sentinel.right, None)
        self.assertIs(intersection.right, mock.sentinel.right)

    def test_point_property(self):
        s_val = 1.0
        t_val = 0.0
        intersection = self._make_one(
            mock.sentinel.left, s_val, mock.sentinel.right, t_val)

        patch = mock.patch(
            'bezier._intersection_helpers._check_close',
            return_value=mock.sentinel.point)
        with patch as mocked:
            self.assertIsNone(intersection._point)
            self.assertIs(intersection.point, mock.sentinel.point)
            self.assertIs(intersection._point, mock.sentinel.point)

            self.assertEqual(mocked.call_count, 1)
            # Make sure the cached value is used on future access.
            self.assertIs(intersection.point, mock.sentinel.point)
            self.assertEqual(mocked.call_count, 1)

    def test_interior_curve_get_unset(self):
        intersection = self._make_one(None, None, None, None)
        self.assertIsNone(intersection._interior_curve)
        with self.assertRaises(AttributeError):
            getattr(intersection, 'interior_curve')

    def test_interior_curve_get_when_set(self):
        intersection = self._make_one(None, None, None, None)
        intersection._interior_curve = mock.sentinel.interior
        self.assertIs(intersection.interior_curve, mock.sentinel.interior)

    def test_interior_curve_set_when_set(self):
        intersection = self._make_one(None, None, None, None)
        intersection.interior_curve = -1
        with self.assertRaises(AttributeError):
            setattr(intersection, 'interior_curve', 0)

    def test_interior_curve_set_valid(self):
        intersection = self._make_one(None, None, None, None)
        intersection.interior_curve = 0
        self.assertEqual(intersection.interior_curve, 0)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

"""py.test shared testing configuration.

This

* gets slow imports out of the way before running tests, so
  as not to have the cost of imports reflected in
  ``py.test --durations=N``.
"""


# pylint: disable=unused-import
from matplotlib import patches  # noqa: F401
from matplotlib import path as _path_mod  # noqa: F401
import matplotlib.pyplot as plt  # noqa: F401
# pylint: enable=unused-import


def pytest_addoption(parser):
    parser.addoption(
        '--ignore-slow',
        dest='ignore_slow',
        action='store_true',
        help='Ignore slow tests.')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper for B |eacute| zier Surfaces / Triangles.

.. |eacute| unicode:: U+000E9 .. LATIN SMALL LETTER E WITH ACUTE
   :trim:

.. testsetup:: *

   import numpy as np
   import bezier
"""


import itertools

from matplotlib import patches
from matplotlib import path as _path_mod
import matplotlib.pyplot as plt
import numpy as np
import six

from bezier import _base
from bezier import _intersection_helpers
from bezier import _surface_helpers
from bezier import curve as _curve_mod


_REPR_TEMPLATE = (
    '<{} (degree={:d}, dimension={:d}, base=({:g}, {:g}), width={:g})>')


class Surface(_base.Base):
    r"""Represents a B |eacute| zier `surface`_.

    .. _surface: https://en.wikipedia.org/wiki/B%C3%A9zier_triangle
    .. _unit simplex:
        https://en.wikipedia.org/wiki/Simplex#The_standard_simplex
    .. _barycentric coordinates:
        https://en.wikipedia.org/wiki/Barycentric_coordinate_system

    We define a B |eacute| zier triangle as a mapping from the
    `unit simplex`_ in 2D (i.e. the unit triangle) onto a surface in an
    arbitrary dimension. We use `barycentric coordinates`_

    .. math::

       \lambda_1 = 1 - s - t, \lambda_2 = s, \lambda_3 = t

    for points in

    .. math::

       \left\{(s, t) \mid 0 \leq s, t, s + t \leq 1\right\}.

    As with curves, using these weights we get convex combinations
    of points :math:`v_{i, j, k}` in some vector space:

    .. math::

       B\left(\lambda_1, \lambda_2, \lambda_3\right) =
           \sum_{i + j + k = d} \binom{d}{i \, j \, k}
           \lambda_1^i \lambda_2^j \lambda_3^k \cdot v_{i, j, k}

    .. note::

       We assume the nodes are ordered from left-to-right and from
       bottom-to-top. So for example, the linear triangle:

       .. code-block:: rest

          (0,0,1)

          (1,0,0)  (0,1,0)

       is ordered as

       .. math::

          \left[\begin{array}{c c c}
              v_{1,0,0} & v_{0,1,0} & v_{0,0,1} \end{array}\right]^T

       the quadratic triangle:

       .. code-block:: rest

          (0,0,2)

          (1,0,1)  (0,1,1)

          (2,0,0)  (1,1,0)  (0,2,0)

       is ordered as

       .. math::

          \left[\begin{array}{c c c c c c}
              v_{2,0,0} & v_{1,1,0} &
              v_{0,2,0} & v_{1,0,1} &
              v_{0,1,1} & v_{0,0,2} \end{array}\right]^T

       the cubic triangle:

       .. code-block:: rest

          (0,0,3)

          (1,0,2)  (0,1,2)

          (2,0,1)  (1,1,1)  (0,2,1)

          (3,0,0)  (2,1,0)  (1,2,0)  (0,3,0)

       is ordered as

       .. math::

          \left[\begin{array}{c c c c c c c c c c}
              v_{3,0,0} & v_{2,1,0} &
              v_{1,2,0} & v_{0,3,0} &
              v_{2,0,1} & v_{1,1,1} &
              v_{0,2,1} & v_{1,0,2} &
              v_{0,1,2} & v_{0,0,3} \end{array}\right]^T

       and so on.

    .. image:: ../images/surface_constructor.png
       :align: center

    .. doctest:: surface-constructor

       >>> import bezier
       >>> nodes = np.array([
       ...     [0.0  , 0.0  ],
       ...     [0.5  , 0.0  ],
       ...     [1.0  , 0.25 ],
       ...     [0.125, 0.5  ],
       ...     [0.375, 0.375],
       ...     [0.25 , 1.0  ],
       ... ])
       >>> surface = bezier.Surface(nodes)
       >>> surface
       <Surface (degree=2, dimension=2)>

    .. testcleanup:: surface-constructor

       import make_images
       make_images.surface_constructor(surface)

    Args:
        nodes (numpy.ndarray): The nodes in the surface. The rows
            represent each node while the columns are the dimension
            of the ambient space.
        base_x (Optional[float]): The :math:`x`-coordinate of the base
           vertex of the sub-triangle that this surface represents.
        base_y (Optional[float]): The :math:`y`-coordinate of the base
           vertex of the sub-triangle that this surface represents.
        width (Optional[float]): The width of the sub-triangle that
           this surface represents.
        _copy (bool): Flag indicating if the nodes should be copied before
            being stored. Defaults to :data:`True` since callers may
            freely mutate ``nodes`` after passing in.
    """

    _area = None
    _edges = None
    _is_valid = None

    def __init__(self, nodes, base_x=0.0, base_y=0.0, width=1.0, _copy=True):
        super(Surface, self).__init__(nodes, _copy=_copy)
        self._base_x = base_x
        self._base_y = base_y
        self._width = width

    def __repr__(self):
        """Representation of current object.

        Returns:
            str: Object representation.
        """
        if self._base_x == 0.0 and self._base_y == 0.0 and self._width == 1.0:
            return super(Surface, self).__repr__()
        else:
            return _REPR_TEMPLATE.format(
                self.__class__.__name__, self.degree, self.dimension,
                self._base_x, self._base_y, self._width)

    @staticmethod
    def _get_degree(num_nodes):
        """Get the degree of the current surface.

        Args:
            num_nodes (int): The number of control points for a
                B |eacute| zier surface.

        Returns:
            int: The degree :math:`d` such that :math:`(d + 1)(d + 2)/2`
            equals ``num_nodes``.

        Raises:
            ValueError: If ``num_nodes`` isn't a triangular number.
        """
        # 8 * num_nodes = 4(d + 1)(d + 2)
        #               = 4d^2 + 12d + 8
        #               = (2d + 3)^2 - 1
        d_float = 0.5 * (np.sqrt(8.0 * num_nodes + 1.0) - 3.0)
        d_int = int(np.round(d_float))
        if (d_int + 1) * (d_int + 2) == 2 * num_nodes:
            return d_int
        else:
            raise ValueError(num_nodes, 'not a triangular number')

    @property
    def area(self):  # pylint: disable=missing-returns-doc
        """float: The area of the current surface.

        Raises:
            NotImplementedError: If the area isn't already cached.
        """
        if self._area is None:
            raise NotImplementedError(
                'Area computation not yet implemented.')
        return self._area

    @property
    def width(self):
        """float: The "width" of the parameterized triangle.

        When re-parameterizing (e.g. via :meth:`subdivide`) we
        specialize the surface from the unit triangle to some
        sub-triangle. After doing this, we re-parameterize so that
        that sub-triangle is treated like the unit triangle.

        To track which sub-triangle we are in during the subdivision
        process, we use the coordinates of the base vertex as well
        as the "width" of each leg.

        .. image:: ../images/surface_width1.png
           :align: center

        .. testsetup:: surface-width1, surface-width2

           import numpy as np
           import bezier

           surface = bezier.Surface(np.array([
               [0.0, 0.0],
               [1.0, 0.0],
               [0.0, 1.0],
           ]))

        .. doctest:: surface-width1

           >>> surface.base_x, surface.base_y
           (0.0, 0.0)
           >>> surface.width
           1.0

        .. testcleanup:: surface-width1

           import make_images
           make_images.surface_width1(surface)

        Upon subdivision, the width halves (and potentially changes sign) and
        the vertex moves to one of four points:

        .. image:: ../images/surface_width2.png
           :align: center

        .. doctest:: surface-width2

           >>> _, sub_surface_b, sub_surface_c, _ = surface.subdivide()
           >>> sub_surface_b.base_x, sub_surface_b.base_y
           (0.5, 0.5)
           >>> sub_surface_b.width
           -0.5
           >>> sub_surface_c.base_x, sub_surface_c.base_y
           (0.5, 0.0)
           >>> sub_surface_c.width
           0.5

        .. testcleanup:: surface-width2

           import make_images
           make_images.surface_width2(sub_surface_b, sub_surface_c)
        """
        return self._width

    @property
    def base_x(self):
        """float: The ``x``-coordinate of the base vertex.

        See :meth:`width` for more detail.
        """
        return self._base_x

    @property
    def base_y(self):
        """float: The ``y``-coordinate of the base vertex.

        See :meth:`width` for more detail.
        """
        return self._base_y

    def _compute_edges(self):
        """Compute the edges of the current surface.

        Returns:
            Tuple[~curve.Curve, ~curve.Curve, ~curve.Curve]: The edges of
            the surface.
        """
        indices1 = slice(0, self.degree + 1)
        indices2 = np.empty(self.degree + 1, dtype=int)
        indices3 = np.empty(self.degree + 1, dtype=int)

        curr2 = self.degree
        curr3 = -1
        for i in six.moves.xrange(self.degree + 1):
            indices2[i] = curr2
            indices3[i] = curr3
            curr2 += self.degree - i
            curr3 -= i + 2

        edge1 = _curve_mod.Curve(self._nodes[indices1, :], _copy=False)
        edge2 = _curve_mod.Curve(self._nodes[indices2, :], _copy=False)
        edge3 = _curve_mod.Curve(self._nodes[indices3, :], _copy=False)
        _surface_helpers.edge_cycle(edge1, edge2, edge3)
        return edge1, edge2, edge3

    def _get_edges(self):
        """Get the edges for the current surface.

        If they haven't been computed yet, first compute and store them.

        This is provided as a means for internal calls to get the edges
        without copying (since :attr:`.edges` copies before giving to
        a user to keep the stored data immutable).

        Returns:
            Tuple[~bezier.curve.Curve, ~bezier.curve.Curve, \
                  ~bezier.curve.Curve]: The edges of
            the surface.
        """
        if self._edges is None:
            self._edges = self._compute_edges()
        return self._edges

    @property
    def edges(self):
        """tuple: The edges of the surface.

        .. doctest:: surface-edges
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.array([
           ...     [0.0   ,  0.0   ],
           ...     [0.5   , -0.1875],
           ...     [1.0   ,  0.0   ],
           ...     [0.1875,  0.5   ],
           ...     [0.625 ,  0.625 ],
           ...     [0.0   ,  1.0   ],
           ... ])
           >>> surface = bezier.Surface(nodes)
           >>> edge1, _, _ = surface.edges
           >>> edge1
           <Curve (degree=2, dimension=2)>
           >>> edge1.nodes
           array([[ 0.  ,  0.    ],
                  [ 0.5 , -0.1875],
                  [ 1.  ,  0.    ]])

        Returns:
            Tuple[~bezier.curve.Curve, ~bezier.curve.Curve, \
                  ~bezier.curve.Curve]: The edges of
            the surface.
        """
        edge1, edge2, edge3 = self._get_edges()
        # NOTE: It is crucial that we return copies here. Since the edges
        #       are cached, if they were mutable, callers could
        #       inadvertently mutate the cached value.
        edge1 = edge1.copy()
        edge2 = edge2.copy()
        edge3 = edge3.copy()
        _surface_helpers.edge_cycle(edge1, edge2, edge3)
        return edge1, edge2, edge3

    def evaluate_barycentric(self, lambda1, lambda2, lambda3):
        r"""Compute a point on the surface.

        Evaluates :math:`B\left(\lambda_1, \lambda_2, \lambda_3\right)`.

        .. image:: ../images/surface_evaluate_barycentric.png
           :align: center

        .. testsetup:: surface-barycentric, surface-barycentric-fail

           import numpy as np
           import bezier
           nodes = np.array([
               [0.0  , 0.0  ],
               [0.5  , 0.0  ],
               [1.0  , 0.25 ],
               [0.125, 0.5  ],
               [0.375, 0.375],
               [0.25 , 1.0  ],
           ])
           surface = bezier.Surface(nodes)

        .. doctest:: surface-barycentric
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.array([
           ...     [0.0  , 0.0  ],
           ...     [0.5  , 0.0  ],
           ...     [1.0  , 0.25 ],
           ...     [0.125, 0.5  ],
           ...     [0.375, 0.375],
           ...     [0.25 , 1.0  ],
           ... ])
           >>> surface = bezier.Surface(nodes)
           >>> point = surface.evaluate_barycentric(0.125, 0.125, 0.75)
           >>> point
           array([ 0.265625 , 0.73046875])

        .. testcleanup:: surface-barycentric

           import make_images
           make_images.surface_evaluate_barycentric(surface, point)

        However, this can't be used for points **outside** the
        reference triangle:

        .. doctest:: surface-barycentric-fail

           >>> surface.evaluate_barycentric(-0.25, 0.75, 0.5)
           Traceback (most recent call last):
             ...
           ValueError: ('Parameters must be positive', -0.25, 0.75, 0.5)

        or for non-Barycentric coordinates;

        .. doctest:: surface-barycentric-fail

           >>> surface.evaluate_barycentric(0.25, 0.25, 0.25)
           Traceback (most recent call last):
             ...
           ValueError: ('Values do not sum to 1', 0.25, 0.25, 0.25)

        Args:
            lambda1 (float): Parameter along the reference triangle.
            lambda2 (float): Parameter along the reference triangle.
            lambda3 (float): Parameter along the reference triangle.

        Returns:
            numpy.ndarray: The point on the surface (as a one dimensional
            NumPy array).

        Raises:
            ValueError: If the weights are not valid barycentric
                coordinates, i.e. they don't sum to ``1``.
            ValueError: If some weights are negative.
        """
        weights_total = lambda1 + lambda2 + lambda3
        if not np.allclose(weights_total, 1.0):
            raise ValueError('Weights do not sum to 1',
                             lambda1, lambda2, lambda3)
        if lambda1 < 0.0 or lambda2 < 0.0 or lambda3 < 0.0:
            raise ValueError('Weights must be positive',
                             lambda1, lambda2, lambda3)

        if self.degree == 1:
            weights = np.array([
                [lambda1, lambda2, lambda3],
            ])
        elif self.degree == 2:
            weights = np.array([
                [
                    lambda1 * lambda1,
                    2.0 * lambda1 * lambda2,
                    lambda2 * lambda2,
                    2.0 * lambda1 * lambda3,
                    2.0 * lambda2 * lambda3,
                    lambda3 * lambda3,
                ]
            ])
        elif self.degree == 3:
            weights = np.array([
                [
                    lambda1 * lambda1 * lambda1,
                    3.0 * lambda1 * lambda1 * lambda2,
                    3.0 * lambda1 * lambda2 * lambda2,
                    lambda2 * lambda2 * lambda2,
                    3.0 * lambda1 * lambda1 * lambda3,
                    6.0 * lambda1 * lambda2 * lambda3,
                    3.0 * lambda2 * lambda2 * lambda3,
                    3.0 * lambda1 * lambda3 * lambda3,
                    3.0 * lambda2 * lambda3 * lambda3,
                    lambda3 * lambda3 * lambda3,
                ]
            ])
        else:
            result = self._nodes
            for reduced_deg in six.moves.xrange(self.degree, 0, -1):
                result = _surface_helpers.de_casteljau_one_round(
                    result, reduced_deg, lambda1, lambda2, lambda3)
            return result.flatten()

        return weights.dot(self._nodes).flatten()  # pylint: disable=no-member

    def evaluate_cartesian(self, s, t):
        r"""Compute a point on the surface.

        Evaluates :math:`B\left(1 - s - t, s, t\right)` by calling
        :meth:`evaluate_barycentric`.

        Args:
            s (float): Parameter along the reference triangle.
            t (float): Parameter along the reference triangle.

        Returns:
            numpy.ndarray: The point on the surface (as a one dimensional
            NumPy array).
        """
        return self.evaluate_barycentric(1.0 - s - t, s, t)

    def evaluate_multi(self, param_vals):
        r"""Compute multiple points on the surface.

        If ``param_vals`` has two columns, this method treats
        them as Cartesian:

        .. image:: ../images/surface_evaluate_multi1.png
           :align: center

        .. doctest:: surface-eval-multi1
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.array([
           ...     [ 0.0, 0.0],
           ...     [ 2.0, 1.0],
           ...     [-3.0, 2.0],
           ... ])
           >>> surface = bezier.Surface(nodes)
           >>> surface
           <Surface (degree=1, dimension=2)>
           >>> param_vals = np.array([
           ...     [0.0  , 0.0  ],
           ...     [0.125, 0.625],
           ...     [0.5  , 0.5  ],
           ... ])
           >>> points = surface.evaluate_multi(param_vals)
           >>> points
           array([[ 0.   , 0.   ],
                  [-1.625, 1.375],
                  [-0.5  , 1.5  ]])

        .. testcleanup:: surface-eval-multi1

           import make_images
           make_images.surface_evaluate_multi1(surface, points)

        and if ``param_vals`` has three columns, treats them as Barycentric:

        .. image:: ../images/surface_evaluate_multi2.png
           :align: center

        .. doctest:: surface-eval-multi2
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.array([
           ...     [ 0. , 0.  ],
           ...     [ 1. , 0.75],
           ...     [ 2. , 1.  ],
           ...     [-1.5, 1.  ],
           ...     [-0.5, 1.5 ],
           ...     [-3. , 2.  ],
           ... ])
           >>> surface = bezier.Surface(nodes)
           >>> surface
           <Surface (degree=2, dimension=2)>
           >>> param_vals = np.array([
           ...     [0.   , 0.25, 0.75 ],
           ...     [1.   , 0.  , 0.   ],
           ...     [0.25 , 0.5 , 0.25 ],
           ...     [0.375, 0.25, 0.375],
           ... ])
           >>> points = surface.evaluate_multi(param_vals)
           >>> points
           array([[-1.75  , 1.75    ],
                  [ 0.    , 0.      ],
                  [ 0.25  , 1.0625  ],
                  [-0.625 , 1.046875]])

        .. testcleanup:: surface-eval-multi2

           import make_images
           make_images.surface_evaluate_multi2(surface, points)

        .. note::

           This currently just uses :meth:`evaluate_cartesian` and
           :meth:`evaluate_barycentric` so is less
           performant than it could be.

        Args:
            param_vals (numpy.ndarray): Array of parameter values (as a
                2D array).

        Returns:
            numpy.ndarray: The point on the surface.

        Raises:
            ValueError: If ``param_vals`` is not a 2D array.
            ValueError: If ``param_vals`` doesn't have 2 or 3 columns.
        """
        if param_vals.ndim != 2:
            raise ValueError('Parameter values must be 2D array')
        num_vals, num_cols = param_vals.shape

        if num_cols == 2:
            transform = self.evaluate_cartesian
        elif num_cols == 3:
            transform = self.evaluate_barycentric
        else:
            raise ValueError(
                'Parameter values must either be Barycentric or Cartesian')

        result = np.empty((num_vals, self.dimension))
        for index in six.moves.xrange(num_vals):
            result[index, :] = transform(*param_vals[index, :])
        return result

    @staticmethod
    def _add_patch(ax, color, edge1, edge2, edge3):
        """Add a polygonal surface patch to a plot.

        Args:
            ax (matplotlib.artist.Artist): A matplotlib axis.
            color (Tuple[float, float, float]): Color as RGB profile.
            edge1 (numpy.ndarray): 2D array of points along a curved edge.
            edge2 (numpy.ndarray): 2D array of points along a curved edge.
            edge3 (numpy.ndarray): 2D array of points along a curved edge.
        """
        # Since the edges overlap, we leave out the first point in each.
        polygon = np.vstack([
            edge1[1:, :],
            edge2[1:, :],
            edge3[1:, :],
        ])
        path = _path_mod.Path(polygon)
        patch = patches.PathPatch(
            path, facecolor=color, alpha=0.6)
        ax.add_patch(patch)

    def plot(self, pts_per_edge, color=None, ax=None,
             with_nodes=False, show=False):
        """Plot the current surface.

        Args:
            pts_per_edge (int): Number of points to plot per edge.
            color (Optional[Tuple[float, float, float]]): Color as RGB profile.
            ax (Optional[matplotlib.artist.Artist]): matplotlib axis object
                to add plot to.
            with_nodes (Optional[bool]): Determines if the control points
                should be added to the plot. Off by default.
            show (Optional[bool]): Flag indicating if the plot should be
                shown.

        Returns:
            matplotlib.artist.Artist: The axis containing the plot. This
            may be a newly created axis.

        Raises:
            NotImplementedError: If the surface's dimension is not ``2``.
        """
        if self.dimension != 2:
            raise NotImplementedError('2D is the only supported dimension',
                                      'Current dimension', self.dimension)

        edge1, edge2, edge3 = self.edges
        s_vals = np.linspace(0.0, 1.0, pts_per_edge)

        points1 = edge1.evaluate_multi(s_vals)
        points2 = edge2.evaluate_multi(s_vals)
        points3 = edge3.evaluate_multi(s_vals)

        if ax is None:
            fig = plt.figure()
            ax = fig.gca()

        line, = ax.plot(points1[:, 0], points1[:, 1], color=color)
        color = line.get_color()
        ax.plot(points2[:, 0], points2[:, 1], color=color)
        ax.plot(points3[:, 0], points3[:, 1], color=color)

        self._add_patch(ax, color, points1, points2, points3)

        if with_nodes:
            ax.plot(self._nodes[:, 0], self._nodes[:, 1],
                    color='black', marker='o', linestyle='None')

        if show:
            plt.show()

        return ax

    def subdivide(self):
        r"""Split the surface into four sub-surfaces.

        Does so by taking the unit triangle (i.e. the domain
        of the surface) and splitting it into four sub-triangles

        .. image:: ../images/surface_subdivide1.png
           :align: center

        Then the surface is re-parameterized via the map to/from the
        given sub-triangles and the unit triangle.

        For example, when a degree two surface is subdivided:

        .. image:: ../images/surface_subdivide2.png
           :align: center

        .. doctest:: surface-subdivide
           :options: +NORMALIZE_WHITESPACE

           >>> nodes = np.array([
           ...     [-1.0 , 0.0 ],
           ...     [ 0.5 , 0.5 ],
           ...     [ 2.0 , 0.0 ],
           ...     [ 0.25, 1.75],
           ...     [ 2.0 , 3.0 ],
           ...     [ 0.0 , 4.0 ],
           ... ])
           >>> surface = bezier.Surface(nodes)
           >>> _, sub_surface_b, _, _ = surface.subdivide()
           >>> sub_surface_b
           <Surface (degree=2, dimension=2, base=(0.5, 0.5), width=-0.5)>
           >>> sub_surface_b.nodes
           array([[ 1.5   , 2.5   ],
                  [ 0.6875, 2.3125],
                  [-0.125 , 1.875 ],
                  [ 1.1875, 1.3125],
                  [ 0.4375, 1.3125],
                  [ 0.5   , 0.25  ]])

        .. testcleanup:: surface-subdivide

           import make_images
           make_images.surface_subdivide1()
           make_images.surface_subdivide2(surface, sub_surface_b)

        Returns:
            Tuple[Surface, Surface, Surface, Surface]: The lower left, central,
            lower right and upper left sub-surfaces (in that order).
        """
        if self.degree == 1:
            # pylint: disable=no-member
            new_nodes = _surface_helpers.LINEAR_SUBDIVIDE.dot(self._nodes)
            # pylint: enable=no-member
            nodes_a = new_nodes[(0, 1, 3), :]
            nodes_b = new_nodes[(4, 3, 1), :]
            nodes_c = new_nodes[(1, 2, 4), :]
            nodes_d = new_nodes[3:, :]
        elif self.degree == 2:
            # pylint: disable=no-member
            new_nodes = _surface_helpers.QUADRATIC_SUBDIVIDE.dot(self._nodes)
            # pylint: enable=no-member
            nodes_a = new_nodes[(0, 1, 2, 5, 6, 9), :]
            nodes_b = new_nodes[(11, 10, 9, 7, 6, 2), :]
            nodes_c = new_nodes[(2, 3, 4, 7, 8, 11), :]
            nodes_d = new_nodes[9:, :]
        elif self.degree == 3:
            # pylint: disable=no-member
            new_nodes = _surface_helpers.CUBIC_SUBDIVIDE.dot(self._nodes)
            # pylint: enable=no-member
            nodes_a = new_nodes[(0, 1, 2, 3, 7, 8, 9, 13, 14, 18), :]
            nodes_b = new_nodes[(21, 20, 19, 18, 16, 15, 14, 10, 9, 3), :]
            nodes_c = new_nodes[(3, 4, 5, 6, 10, 11, 12, 16, 17, 21), :]
            nodes_d = new_nodes[18:, :]
        elif self.degree == 4:
            # pylint: disable=no-member
            new_nodes = _surface_helpers.QUARTIC_SUBDIVIDE.dot(self._nodes)
            # pylint: enable=no-member
            nodes_a = new_nodes[
                (0, 1, 2, 3, 4, 9, 10, 11, 12, 17, 18, 19, 24, 25, 30), :]
            nodes_b = new_nodes[
                (34, 33, 32, 31, 30, 28, 27, 26, 25, 21, 20, 19, 13, 12, 4), :]
            nodes_c = new_nodes[
                (4, 5, 6, 7, 8, 13, 14, 15, 16, 21, 22, 23, 28, 29, 34), :]
            nodes_d = new_nodes[30:, :]
        else:
            nodes_a = _surface_helpers.specialize_surface(
                self._nodes, self.degree,
                (1.0, 0.0, 0.0), (0.5, 0.5, 0.0), (0.5, 0.0, 0.5))
            nodes_b = _surface_helpers.specialize_surface(
                self._nodes, self.degree,
                (0.0, 0.5, 0.5), (0.5, 0.0, 0.5), (0.5, 0.5, 0.0))
            nodes_c = _surface_helpers.specialize_surface(
                self._nodes, self.degree,
                (0.5, 0.5, 0.0), (0.0, 1.0, 0.0), (0.0, 0.5, 0.5))
            nodes_d = _surface_helpers.specialize_surface(
                self._nodes, self.degree,
                (0.5, 0.0, 0.5), (0.0, 0.5, 0.5), (0.0, 0.0, 1.0))

        half_width = 0.5 * self._width
        shifted_x = self._base_x + half_width
        shifted_y = self._base_y + half_width
        return (
            Surface(nodes_a, base_x=self._base_x, base_y=self._base_y,
                    width=half_width, _copy=False),
            Surface(nodes_b, base_x=shifted_x, base_y=shifted_y,
                    width=-half_width, _copy=False),
            Surface(nodes_c, base_x=shifted_x, base_y=self._base_y,
                    width=half_width, _copy=False),
            Surface(nodes_d, base_x=self._base_x, base_y=shifted_y,
                    width=half_width, _copy=False),
        )

    def _compute_valid(self):
        r"""Determines if the current surface is "valid".

        Does this by checking if the Jacobian of the map from the
        reference triangle is nonzero.

        Returns:
            bool: Flag indicating if the current surface is valid.

        Raises:
            NotImplementedError: If the degree is greater than 3.
            NotImplementedError: If the surface is quadratic or cubic
                but in dimension other than :math:`\mathbf{R}^2`.
        """
        if self.degree == 1:
            # In the linear case, we are only invalid if the points
            # are collinear.
            # pylint: disable=no-member
            first_deriv = self._nodes[1:, :] - self._nodes[:-1, :]
            # pylint: enable=no-member
            return np.linalg.matrix_rank(first_deriv) == 2
        elif self.degree in (2, 3):
            if self.dimension != 2:
                raise NotImplementedError(
                    'Cubic/quadratic validity check only implemented in R^2')

            if self.degree == 2:
                bernstein = _surface_helpers.quadratic_jacobian_polynomial(
                    self._nodes)
            else:
                bernstein = _surface_helpers.cubic_jacobian_polynomial(
                    self._nodes)

            # Form the polynomial p(s, t) as a Surface.
            jac_poly = Surface(bernstein, _copy=False)
            # Find the sign of the polynomial, where 0 means mixed.
            poly_sign = _surface_helpers.polynomial_sign(jac_poly)
            return poly_sign != 0
        else:
            raise NotImplementedError(
                'Degrees 1, 2 and 3 only supported at this time')

    @property
    def is_valid(self):
        """bool: Flag indicating if the surface is "valid".

        Here, "valid" means there are no self-intersections or
        singularities.

        This checks if the Jacobian of the map from the reference
        triangle is nonzero. For example, a linear "surface"
        with collinear points is invalid:

        .. image:: ../images/surface_is_valid1.png
           :align: center

        .. doctest:: surface-is-valid1

           >>> nodes = np.array([
           ...     [0.0, 0.0],
           ...     [1.0, 1.0],
           ...     [2.0, 2.0],
           ... ])
           >>> surface = bezier.Surface(nodes)
           >>> surface.is_valid
           False

        .. testcleanup:: surface-is-valid1

           import make_images
           make_images.surface_is_valid1(surface)

        while a quadratic surface with one straight side:

        .. image:: ../images/surface_is_valid2.png
           :align: center

        .. doctest:: surface-is-valid2

           >>> nodes = np.array([
           ...     [ 0.0  , 0.0  ],
           ...     [ 0.5  , 0.125],
           ...     [ 1.0  , 0.0  ],
           ...     [-0.125, 0.5  ],
           ...     [ 0.5  , 0.5  ],
           ...     [ 0.0  , 1.0  ],
           ... ])
           >>> surface = bezier.Surface(nodes)
           >>> surface.is_valid
           True

        .. testcleanup:: surface-is-valid2

           import make_images
           make_images.surface_is_valid2(surface)

        though not all higher degree surfaces are valid:

        .. image:: ../images/surface_is_valid3.png
           :align: center

        .. doctest:: surface-is-valid3

           >>> nodes = np.array([
           ...     [1.0, 0.0],
           ...     [0.0, 0.0],
           ...     [1.0, 1.0],
           ...     [0.0, 0.0],
           ...     [0.0, 0.0],
           ...     [0.0, 1.0],
           ... ])
           >>> surface = bezier.Surface(nodes)
           >>> surface.is_valid
           False

        .. testcleanup:: surface-is-valid3

           import make_images
           make_images.surface_is_valid3(surface)
        """
        if self._is_valid is None:
            self._is_valid = self._compute_valid()
        return self._is_valid

    def locate(self, point):
        r"""Find a point on the current surface.

        Solves for :math:`s` and :math:`t` in :math:`B(s, t) = p`.

        .. note::

           A unique solution is only guaranteed if the current surface is
           valid. This code assumes a valid surface, but doesn't check.

        .. image:: ../images/surface_locate.png
           :align: center

        .. doctest:: surface-locate

           >>> surface = bezier.Surface(np.array([
           ...     [0.0 ,  0.0 ],
           ...     [0.5 , -0.25],
           ...     [1.0 ,  0.0 ],
           ...     [0.25,  0.5 ],
           ...     [0.75,  0.75],
           ...     [0.0 ,  1.0 ],
           ... ]))
           >>> point = np.array([[0.59375, 0.25]])
           >>> s, t = surface.locate(point)
           >>> s
           0.5
           >>> t
           0.25

        .. testcleanup:: surface-locate

           import make_images
           make_images.surface_locate(surface, point)

        Args:
            point (numpy.ndarray): A (``1xD``) point on the surface,
                where :math:`D` is the dimension of the surface.

        Returns:
            Optional[Tuple[float, float]]: The :math:`s` and :math:`t`
            values corresponding to ``x_val`` and ``y_val`` or
            :data:`None` if the point is not on the ``surface``.

        Raises:
            NotImplementedError: If the surface isn't in :math:`\mathbf{R}^2`.
            ValueError: If the dimension of the ``point`` doesn't match the
                dimension of the current surface.
        """
        if self.dimension != 2:
            raise NotImplementedError('Only 2D surfaces supported.')

        if point.shape != (1, self.dimension):
            raise ValueError('Point is not in same dimension as surface',
                             point, 'Shape expected:', (1, self.dimension))

        return _surface_helpers.locate_point(self, point[0, 0], point[0, 1])

    def intersect(self, other):
        """Find the common intersection with another surface.

        Args:
            other (Surface): Other surface to intersect with.

        Returns:
            list: List of intersection objects (possibly empty).

        Raises:
            TypeError: If ``other`` is not a surface.
            NotImplementedError: If both surfaces aren't two-dimensional.
        """
        if not isinstance(other, Surface):
            raise TypeError('Can only intersect with another surface',
                            'Received', other)
        if self.dimension != 2 or other.dimension != 2:
            raise NotImplementedError(
                'Intersection only implemented in 2D')

        edges1 = self._get_edges()
        lin1 = six.moves.map(
            _intersection_helpers.Linearization.from_shape,
            edges1)
        edges2 = other._get_edges()  # pylint: disable=protected-access
        lin2 = six.moves.map(
            _intersection_helpers.Linearization.from_shape,
            edges2)

        edge_pairs = itertools.product(lin1, lin2)
        intersections = _intersection_helpers.all_intersections(
            edge_pairs)

        # Classify each intersection.
        duplicates = []
        uniques = []
        for intersection in intersections:
            changed = _surface_helpers.handle_corners(intersection)
            if changed:
                duplicates.append(intersection)
            else:
                interior = _surface_helpers.classify_intersection(
                    intersection)
                intersection.interior_curve = interior
                uniques.append(intersection)

        # NOTE: This is an extra check to make sure we found duplicate
        #       intersections at corners. It is not strictly needed,
        #       it is just provided as a sanity check.
        _surface_helpers.verify_duplicates(duplicates, uniques)
        return _surface_helpers.combine_intersections(
            uniques, self, other)

    def elevate(self):
        r"""Return a degree-elevated version of the current surface.

        Does this by converting the current nodes
        :math:`\left\{v_{i, j, k}\right\}_{i + j + k = d}` to new nodes
        :math:`\left\{w_{i, j, k}\right\}_{i + j + k = d + 1}`. Does so
        by re-writing

        .. math::

           E\left(\lambda_1, \lambda_2, \lambda_3\right) =
               \left(\lambda_1 + \lambda_2 + \lambda_3\right)
               B\left(\lambda_1, \lambda_2, \lambda_3\right) =
               \sum_{i + j + k = d + 1} \binom{d + 1}{i \, j \, k}
               \lambda_1^i \lambda_2^j \lambda_3^k \cdot w_{i, j, k}

        In this form, we must have

        .. math::

           \begin{align*}
           \binom{d + 1}{i \, j \, k} \cdot w_{i, j, k} &=
               \binom{d}{i - 1 \, j \, k} \cdot v_{i - 1, j, k} +
               \binom{d}{i \, j - 1 \, k} \cdot v_{i, j - 1, k} +
               \binom{d}{i \, j \, k - 1} \cdot v_{i, j, k - 1} \\
           \Longleftrightarrow (d + 1) \cdot w_{i, j, k} &=
               i \cdot v_{i - 1, j, k} + j \cdot v_{i, j - 1, k} +
               k \cdot v_{i, j, k - 1}
           \end{align*}

        where we assume that, for example, :math:`v_{i, j, k - 1}` is
        :math:`0` (or any other unused value) if :math:`k = 0`.

        .. doctest:: surface-elevate
           :options: +NORMALIZE_WHITESPACE

           >>> surface = bezier.Surface(np.array([
           ...     [0.0, 0.0],
           ...     [1.0, 0.0],
           ...     [0.0, 1.0],
           ... ]))
           >>> surface
           <Surface (degree=1, dimension=2)>
           >>> new_surface = surface.elevate()
           >>> new_surface
           <Surface (degree=2, dimension=2)>
           >>> new_surface.nodes
           array([[ 0. , 0. ],
                  [ 0.5, 0. ],
                  [ 1. , 0. ],
                  [ 0. , 0.5],
                  [ 0.5, 0.5],
                  [ 0. , 1. ]])

        Returns:
            Surface: The degree-elevated surface.
        """
        num_nodes, _ = self._nodes.shape
        # (d + 1)(d + 2)/2 --> (d + 2)(d + 3)/2
        num_new = num_nodes + self.degree + 2
        new_nodes = np.zeros((num_new, self.dimension))

        # NOTE: We start from the index triples (i, j, k) for the current
        #       nodes and map them onto (i + 1, j, k), etc. This index
        #       tracking is also done in :func:`.de_casteljau_one_round`.
        index = 0
        # parent_i1 = index + k
        # parent_i2 = index + k + 1
        # parent_i3 = index + degree + 2
        parent_i1 = 0
        parent_i2 = 1
        parent_i3 = self.degree + 2
        for k in six.moves.xrange(self.degree + 1):
            for j in six.moves.xrange(self.degree + 1 - k):
                i = self.degree - j - k
                new_nodes[parent_i1, :] += (i + 1) * self._nodes[index, :]
                new_nodes[parent_i2, :] += (j + 1) * self._nodes[index, :]
                new_nodes[parent_i3, :] += (k + 1) * self._nodes[index, :]
                # Update all the indices.
                parent_i1 += 1
                parent_i2 += 1
                parent_i3 += 1
                index += 1

            # Update the indices that depend on k.
            parent_i1 += 1
            parent_i2 += 1

        # Hold off on division until the end, to (attempt to) avoid round-off.
        denominator = self.degree + 1.0
        new_nodes /= denominator

        return Surface(new_nodes, _copy=False)



# See the License for the specific language governing permissions and
# limitations under the License.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import numpy as np

from tests import utils


class TestBase(utils.NumPyTestCase):

    @staticmethod
    def _get_target_class():
        from bezier import _base

        return _base.Base

    def _make_one(self, *args, **kwargs):
        klass = self._get_target_class()
        return klass(*args, **kwargs)

    def test_constructor(self):
        nodes = np.array([
            [0.0, 0.0],
            [1.0, 1.0],
            [2.0, 3.0],
        ])
        shape = self._make_one(nodes)
        self.assertEqual(shape._degree, 3)
        self.assertEqual(shape._dimension, 2)
        self.assertIsNot(shape._nodes, nodes)
        self.assertEqual(shape._nodes, nodes)

    def test_constructor_without_copy(self):
        nodes = np.array([
            [0.0, 0.0],
            [1.0, 1.0],
            [2.0, 3.0],
        ])
        shape = self._make_one(nodes, _copy=False)
        self.assertEqual(shape._degree, 3)
        self.assertEqual(shape._dimension, 2)
        self.assertIs(shape._nodes, nodes)

    def test_constructor_wrong_dimension(self):
        nodes = np.array([1.0, 2.0])
        with self.assertRaises(ValueError):
            self._make_one(nodes)

        nodes = np.zeros((2, 2, 2))
        with self.assertRaises(ValueError):
            self._make_one(nodes)

    def test_constructor_insufficient_nodes(self):
        nodes = np.zeros((0, 2))
        with self.assertRaises(ValueError):
            self._make_one(nodes)

    def test__get_degree(self):
        klass = self._get_target_class()
        num_nodes = 41
        self.assertEqual(num_nodes, klass._get_degree(num_nodes))

    def test_degree_property(self):
        degree = 6
        nodes = np.zeros((degree, 2))
        shape = self._make_one(nodes)
        self.assertEqual(shape.degree, degree)
        self.assertEqual(shape._degree, degree)

    def test_dimension_property(self):
        dimension = 4
        nodes = np.zeros((3, dimension))
        shape = self._make_one(nodes)
        self.assertEqual(shape.dimension, dimension)
        self.assertEqual(shape._dimension, dimension)

    def test_nodes_property(self):
        nodes = np.array([
            [0.0, 0.0],
            [1.0, 2.0],
        ])
        shape = self._make_one(nodes)
        self.assertEqual(shape.nodes, nodes)
        self.assertIsNot(shape.nodes, nodes)

    def test_copy(self):
        np_shape = (2, 2)
        shape = self._make_one(np.zeros(np_shape))
        fake_nodes = mock.Mock(ndim=2, shape=np_shape)
        shape._nodes = fake_nodes

        copied_nodes = np.zeros(np_shape)
        fake_nodes.copy.return_value = copied_nodes

        new_shape = shape.copy()
        self.assertIsInstance(new_shape, self._get_target_class())
        self.assertIs(new_shape._nodes, copied_nodes)

        fake_nodes.copy.assert_called_once_with()

    def test___eq__(self):
        nodes1 = np.zeros((4, 3))
        shape1 = self._make_one(nodes1)
        nodes2 = np.zeros((4, 3))
        shape2 = self._make_one(nodes2)
        self.assertEqual(shape1, shape2)

    def test___ne__different_degree(self):
        nodes1 = np.zeros((1, 3))
        shape1 = self._make_one(nodes1)
        nodes2 = np.zeros((4, 3))
        shape2 = self._make_one(nodes2)
        self.assertNotEqual(shape1, shape2)

    def test___ne__different_dimension(self):
        nodes1 = np.zeros((4, 2))
        shape1 = self._make_one(nodes1)
        nodes2 = np.zeros((4, 3))
        shape2 = self._make_one(nodes2)
        self.assertNotEqual(shape1, shape2)

    def test___ne__different_class(self):
        nodes = np.zeros((4, 2))

        base_class = self._get_target_class()
        new_class1 = type('NewOne', (base_class,), {})
        new_class2 = type('NewTwo', (base_class,), {})

        shape1 = new_class1(nodes)
        shape2 = new_class2(nodes)
        self.assertNotEqual(shape1, shape2)

    def test___repr__(self):
        nodes = np.zeros((4, 3))
        shape = self._make_one(nodes)
        expected = '<Base (degree=4, dimension=3)>'
        self.assertEqual(repr(shape), expected)
